from pyha.common.hwsim import HW
from pyha.common.sfix import resize, scalb, ComplexSfix, Sfix


# blade rf format to 18 bit format and complex
# idea is to turn (4 downto -11) -> (0 downto -17)
class BladeToComplex(HW):
    def __init__(self):
        self.out = ComplexSfix()

        self._delay = 1

    def main(self, i, q):
        outi = resize(scalb(i, 4), 0, -17)
        outq = resize(scalb(q, 4), 0, -17)

        # make complex
        self.next.out = ComplexSfix(outi, outq)
        return self.out

    def model_main(self, i, q):
        return i * (2 ** 4) + q * (2 ** 4) * 1j


# move back to blade format
class NormalToBlade(HW):
    def __init__(self):
        self.out = Sfix()

        self._delay = 1

    def main(self, x):
        self.next.out = resize(scalb(x, -4), 0, -15)
        return self.out

    def model_main(self, x):
        return x * (2 ** -4)


import collections

import numpy as np
import scipy
from scipy import signal


def escape_for_vhdl(x: str) -> str:
    vhdl_reserved_names = ['abs', 'after', 'alias', 'all', 'and', 'architecture',
                           'array', 'assert', 'attribute', 'begin', 'block', 'body',
                           'buffer', 'bus', 'case', 'component', 'configuration',
                           'constant', 'disconnect', 'downto', 'else', 'elsif', 'end',
                           'entity', 'exit', 'file', 'for', 'function', 'generate', 'generic',
                           'group', 'guarded', 'if', 'impure', 'in', 'inertial', 'inout', 'is',
                           'label', 'library', 'linkage', 'literal', 'loop', 'map', 'mod',
                           'nand', 'new', 'next', 'nor', 'not', 'null', 'of', 'on', 'open', 'or',
                           'others', 'out', 'package', 'port', 'postponed', 'procedure',
                           'process', 'pure', 'range', 'record', 'register', 'reject', 'rem',
                           'report', 'return', 'rol', 'ror', 'select', 'severity', 'signal',
                           'shared', 'sla', 'sll', 'sra', 'srl', 'subtype', 'then', 'to',
                           'transport', 'type', 'unaffected', 'units', 'until', 'use',
                           'variable', 'wait', 'when', 'while', 'with', 'xnor', 'xor']

    if x.lower() in vhdl_reserved_names or x[0] == '_':
        return f'\\{x}\\'  # "escape" reserved name
    return x


def is_power2(num):
    # Author: A.Polino
    'states if a number is a power of two'
    return num != 0 and ((num & (num - 1)) == 0)


def get_iterable(x):
    if isinstance(x, collections.Iterable):
        return x
    else:
        return [x]


def tabber(str):
    TAB = '    '
    """ Add tab infront of every line """
    # if x is '' then dont add tabs, because much used textwrap.dedent deletes empty tabs...so unit tests will be unhappy
    return '\n'.join([f'{TAB}{x}' if x != '' else x for x in str.splitlines()])


def load_gnuradio_file(file: str):
    return scipy.fromfile(open(file), dtype=scipy.complex64)


def resample_gnuradio(file: str, ratio: float):
    """ ratio is current_fs / desired_fs """
    f = load_gnuradio_file(file)
    resampled = signal.resample(f, len(f) * ratio)
    return resampled


def save_gnuradio_file(file: str, arr):
    conv = np.array(arr).astype(scipy.complex64)
    conv.tofile(file)


def hex_to_bitstr(hstr):
    """ http://stackoverflow.com/questions/1425493/convert-hex-to-binary """
    if isinstance(hstr, int):
        hstr = hex(hstr)
    if hstr[0:2] in ('0x', '0X'):
        hstr = hstr[2:]
    my_hexdata = hstr
    scale = 16  ## equals to hexadecimal
    num_of_bits = int(len(my_hexdata) * np.log2(scale))
    return bin(int(my_hexdata, scale))[2:].zfill(num_of_bits)


def hex_to_bool_list(hstr):
    return [bool(int(x)) for x in hex_to_bitstr(hstr)]

def int_to_bool_list(hstr):
    return [bool(int(x)) for x in hex_to_bitstr(hstr)]


def test_hex_to_bits():
    assert hex_to_bitstr('FF') == '11111111'
    assert hex_to_bitstr('0xFF') == '11111111'
    assert hex_to_bitstr('0XFF') == '11111111'
    assert hex_to_bitstr('0F') == '00001111'

    assert hex_to_bool_list('F') == [True, True, True, True]
    assert hex_to_bool_list('0F') == [False, False, False, False, True, True, True, True]

    assert hex_to_bool_list(0xF) == [True, True, True, True]
    assert hex_to_bool_list(0xFFFF) == [True] * 16


def bools_to_hex(bl):
    bitstr = bools_to_bitstr(bl)
    return hex(int(bitstr, 2))

def bools_to_bitstr(bl):
    return ''.join(str(int(x)) for x in bl)

def test_bools_to_hex():
    assert bools_to_hex([True, True, True, True]) == '0xf'
    assert bools_to_hex([True, True]) == '0x3'
    assert bools_to_hex([False, True, True, True, True, False, True, True]) == '0x7b'
    assert bools_to_hex([0, 1, 1, 1, 1, 0, 1, 1]) == '0x7b'



# TODO: This file is 100% mess, only works thanks to unit tests
from enum import Enum

from redbaron import GetitemNode, DefNode, AssignmentNode, IntNode, NameNode, CallArgumentNode
from redbaron.nodes import DefArgumentNode, AtomtrailersNode

from pyha.common.hwsim import HW
from pyha.common.sfix import Sfix, ComplexSfix
from pyha.common.util import escape_for_vhdl
from pyha.conversion.extract_datamodel import DataModel


class ExceptionCoupling(Exception):
    pass


def pytype_to_vhdl(var):
    if type(var) is bool:
        return 'boolean'
    elif type(var) is int:
        return 'integer'
    elif type(var) is float:
        return 'real'
    elif type(var) is Sfix:
        return f'sfixed({var.left} downto {var.right})'
    elif type(var) is ComplexSfix:
        left, right = bounds_to_str(var)
        return f'complex_sfix{left}{right}'
    elif type(var) is list:
        arr_token = f'_list_t(0 to {len(var) - 1})'
        if type(var[0]) is bool:
            return 'boolean' + arr_token
        elif type(var[0]) is int:
            return 'integer' + arr_token
        elif type(var[0]) is float:
            return 'real' + arr_token
        elif type(var[0]) is Sfix:
            left, right = bounds_to_str(var[0])
            return f'sfixed{left}{right}{arr_token}'
        elif type(var[0]) is ComplexSfix:
            left, right = bounds_to_str(var[0])
            return f'complex_sfix{left}{right}{arr_token}'
        elif isinstance(var[0], HW):
            return pytype_to_vhdl(var[0]) + arr_token
        else:
            assert 0
    elif isinstance(var, HW):
        idstr = get_instance_vhdl_name(var)
        return idstr
    elif isinstance(var, Enum):
        return type(var).__name__
    else:
        assert 0


def bounds_to_str(var):
    left = var.left if var.left >= 0 else '_' + str(abs(var.left))
    right = var.right if var.right >= 0 else '_' + str(abs(var.right))
    return left, right


def reset_maker(self_data, recursion_depth=0):
    lines = []
    prefix = 'self' if recursion_depth == 0 else ''
    for var_name, var_value in self_data.items():
        if var_name == 'next':
            continue
        var_name = escape_for_vhdl(var_name)

        if isinstance(var_value, (Sfix, ComplexSfix)):
            lines.append(f'{prefix}.\\next\\.{var_name} := {var_value.vhdl_reset()};')

        elif isinstance(var_value, (Enum)):
            lines.append(f'{prefix}.\\next\\.{var_name} := {var_value.name};')

        # list of submodules
        elif isinstance(var_value, list) and isinstance(var_value[0], HW):
            for i, x in enumerate(var_value):
                dm = DataModel(x)
                resets = reset_maker(dm.self_data, recursion_depth + 1)  # recursion here
                vars = [f'{prefix}.{var_name}({i}){x}' for x in resets]
                lines.extend(vars)

        # some other list
        elif isinstance(var_value, list):
            lines.append(f'{prefix}.\\next\\.' + list_reset('', var_name, var_value))

        # submodule
        elif isinstance(var_value, HW):
            if recursion_depth == 0:
                lines.append(f'{get_instance_vhdl_name(var_value)}.\\_pyha_reset_self\\(self.{var_name});')
            else:
                # submodule of lists of submodules
                dm = DataModel(var_value)
                resets = reset_maker(dm.self_data, recursion_depth + 1)  # recursion here
                vars = [f'{prefix}.{var_name}{x}' for x in resets]
                lines.extend(vars)
        else:
            lines.append(f'{prefix}.\\next\\.{var_name} := {var_value};')

    return lines


def list_reset(prefix, key, value):
    if isinstance(value[0], (Sfix, ComplexSfix)):
        lstr = '(' + ', '.join(x.vhdl_reset() for x in value) + ')'
    else:
        lstr = '(' + ', '.join(str(x) for x in value) + ')'
    tmp = f'{prefix + key} := {lstr};'
    return tmp


def get_instance_vhdl_name(variable=None, name: str = '', id: int = 0):
    if variable is not None:
        name = type(variable).__name__
        id = variable._pyha_instance_id
    return escape_for_vhdl(f'{name}_{id}')


class VHDLType:
    """ This merges converter and datamodel code. Converter provides a variable,
    datamodel provides a type, this finds a type fof variable"""

    _datamodel = None

    @classmethod
    def set_datamodel(cls, dm: DataModel):
        cls._datamodel = dm

    @classmethod
    def get_reset(cls):
        return reset_maker(cls._datamodel.self_data)

    @classmethod
    def get_self_vhdl_name(cls):
        if cls._datamodel.obj is None:
            return 'unknown_name'
        return get_instance_vhdl_name(cls._datamodel.obj)

    @classmethod
    def get_constants(cls):
        if cls._datamodel is None:
            return []
        ret = []
        for k, v in cls._datamodel.constants.items():
            t = VHDLType(tuple_init=(k, v))
            ret.append(t)
        return ret

    @classmethod
    def get_self(cls):
        if cls._datamodel is None:
            return []
        ret = []
        for k, v in cls._datamodel.self_data.items():
            if k == 'next':
                continue
            t = VHDLType(tuple_init=(k, v))

            # todo remove this hack
            if isinstance(v, HW):
                t.var_type += '.self_t'
            ret.append(t)
        return ret

    @classmethod
    def get_complex_vars(cls):
        typedefs = cls._get_vars_by_type(ComplexSfix)
        return typedefs

    @classmethod
    def _get_vars_by_type(cls, find_type):
        def scan_arr(arr, find_type):
            ret = []
            for var in arr:
                if isinstance(var, find_type):
                    ret.append(var)
                elif isinstance(var, list) and isinstance(var[0], find_type):
                    ret.append(var[0])
            return ret

        vars = []

        # from self.data
        vars.extend(scan_arr(cls._datamodel.self_data.values(), find_type))

        # from constants
        vars.extend(scan_arr(cls._datamodel.constants.values(), find_type))

        # from locals
        for func in cls._datamodel.locals.values():
            vars.extend(scan_arr(func.values(), find_type))

        return vars

    @classmethod
    def get_typedef_vars(cls):
        """ Return all variables that require new type definition in VHDL, for example arrays"""
        typedefs = cls._get_vars_by_type(list)

        # ignore boolean arrays, there is global definition for that
        typedefs = [x for x in typedefs if not isinstance(x[0], bool)]
        return typedefs

    @classmethod
    def get_enum_vars(cls):
        typedefs = cls._get_vars_by_type(Enum)
        return typedefs

    def __init__(self, name=None, red_node=None, var_type: str = None, port_direction: str = None, value=None,
                 tuple_init=None):
        from pyha.conversion.converter import NameNodeConv
        self.value = value
        self.red_node = red_node
        self.port_direction = port_direction
        self.variable = None
        if tuple_init is not None:
            self.name = escape_for_vhdl(tuple_init[0])
            self.variable = tuple_init[1]
            self.var_type = pytype_to_vhdl(self.variable)
            # self.var_typedef = self.deduce_typedef(tuple_init[1])
            return

            # hardcoded types
        self.var_type = var_type
        if isinstance(name, NameNodeConv):
            self.name = name
        else:
            assert isinstance(name, str)
            self.name = escape_for_vhdl(name)

        if str(name) == 'self':
            self.var_type = 'self_t'
            self.port_direction = 'inout'
            return

        # hack to make 'self.target.name' duck typing work
        class Hack:
            def __init__(self, name):
                self.name = name

        self.target = Hack(self.name)

        if self.var_type is None:
            self.var_type = self.deduce_type()

    def __str__(self):
        var_type = self.var_type or 'unknown_type'
        port_direction = self.port_direction or ''
        default_value = f':={self.value}' if self.value else ''
        tmp_str = f'{self.name}:{port_direction} {var_type}{default_value}'
        return tmp_str

    def __repr__(self):
        return self.__str__()

    def deduce_type(self):
        if self._datamodel is None:  # support converter tests
            return self.var_type or 'unknown_type'

        self.variable = self.get_var_datamodel()

        type = pytype_to_vhdl(self.variable)
        return type

    def get_var_datamodel(self):
        # is 'self.x' something
        if isinstance(self.red_node, AtomtrailersNode) and str(self.red_node[0]) == 'self':
            var = self.walk_self_data(self.red_node)
        elif isinstance(self.red_node, IntNode):  # integer constant
            var = int(self.red_node.value)
        # boolean constant
        elif isinstance(self.red_node, NameNode) and self.red_node.value in ['False', 'True']:
            var = bool(self.red_node.value)
        # Sfix constant
        elif isinstance(self.red_node, AtomtrailersNode) and self.red_node.value[0] == 'Sfix':
            raise Exception('Return Sfix constant? Not allowed because too hard to implement :O')
        else:  # dealing with locals (includes all arguments!)
            name = self._real_name()
            var = self._datamodel.locals[self._defined_in_function()][name]
            try:
                if isinstance(self.red_node[1], GetitemNode):
                    try:
                        index = int(self.red_node.value[1].value.value)
                    except AttributeError:
                        index = int(self.red_node.value[1].value)
                    var = var[index]
            except:
                pass
        return var

    def walk_self_data(self, atom_trailer):
        """ atom_trailer is something like this: self.a.b.c.d
            This finds type of such nested variable
        """

        def find_from_self(atom_trailer):
            var = self._datamodel.self_data
            for x in atom_trailer[1:]:
                if str(x) == 'next': continue
                if not isinstance(x, GetitemNode):
                    var = var[str(x)]
                else:
                    # index is some variable -> just take first element
                    if isinstance(x.value, NameNode):
                        var = var[0]
                    else:
                        var = var[int(str(x.value))]
            return var

        def find_from_const(atom_trailer):
            var = self._datamodel.constants
            for x in atom_trailer[1:]:
                if str(x) == 'next': continue
                if not isinstance(x, GetitemNode):
                    var = var[str(x)]
                else:
                    # index is some variable -> just take first element
                    if isinstance(x.value, NameNode):
                        var = var[0]
                    else:
                        var = var[int(str(x.value))]
            return var

        try:
            var = find_from_self(atom_trailer)
        except KeyError:
            var = find_from_const(atom_trailer)
        return var

    def _defined_in_function(self):
        defn = self.red_node
        while type(defn) is not DefNode:
            defn = defn.parent
        return defn.name

    def _real_name(self):
        # VHDLType.name could be something else, like 'ret_0'
        if isinstance(self.red_node, DefArgumentNode):
            name = str(self.red_node.target)
        elif isinstance(self.red_node, AtomtrailersNode):
            # if len(self.red_node('call')):
            #     arr = self.red_node.find('getitem')
            #     if arr:
            #         return str(arr.previous)
            #     else:
            #         return self.name
            getitem = self.red_node.find('getitem')
            if getitem is not None:
                return str(getitem.previous)
            name = str(self.red_node)
        elif isinstance(self.red_node, AssignmentNode):
            name = str(self.red_node.target)
        elif isinstance(self.red_node, CallArgumentNode):
            getitem = self.red_node.find('getitem')
            if getitem is not None:
                return str(getitem.previous)
            name = str(self.red_node.value)
        else:
            name = str(self.red_node)

        return name


class VHDLVariable(VHDLType):
    def __str__(self):
        if self.name == 'self':
            # without this result would be 'variable self: inout self_t;'
            return 'variable self: self_t;'
        sup = super().__str__()
        return 'variable ' + sup + ';'



import logging
import os
import subprocess
from pathlib import Path

import pyha
from pyha import common
from pyha.conversion.conversion import Conversion
from pyha.simulation.cocotb import CocotbAuto


class SimProvider:
    def __init__(self, base_path, model, sim_type):
        self.logger = logging.getLogger(__name__)
        self.sim_type = sim_type
        self.base_path = base_path

        self.src_path = self.base_path / 'src'
        self.quartus_path = self.base_path / 'quartus'

        if not self.src_path.exists():
            os.makedirs(self.src_path)

        if not self.quartus_path.exists():
            os.makedirs(self.quartus_path)


        self.model = model
        self.conv = Conversion(self.model)

    def get_conversion_sources(self):
        src = [pyha.__path__[0] + '/common/hdl/pyha_util.vhd']
        src += self.conv.write_vhdl_files(self.src_path)
        return src

    def main(self):
        from pyha.simulation.simulation_interface import SIM_GATE
        src = self.get_conversion_sources()
        if self.sim_type == SIM_GATE:
            self.make_quartus_project()
            vho = self.make_quartus_netlist()
            src = [str(vho)]
            # src = ['/home/gaspar/git/pyha/playground/conv/simulation/modelsim/quartus_project.vho']

        return CocotbAuto(self.base_path, src, self.conv.outputs)

    def make_quartus_project(self):
        rules = {}
        rules['DEVICE'] = 'EP4CE40F23C8'
        rules['TOP_LEVEL_ENTITY'] = 'top'
        rules['PROJECT_OUTPUT_DIRECTORY'] = 'output_files'

        # EDA generation not working without this
        rules['EDA_SIMULATION_TOOL'] = '"ModelSim-Altera (VHDL)"'

        rules['EDA_GENERATE_FUNCTIONAL_NETLIST'] = 'ON -section_id eda_simulation'
        rules['VHDL_INPUT_VERSION'] = 'VHDL_2008'
        rules['VHDL_SHOW_LMF_MAPPING_MESSAGES'] = 'OFF'

        common_path = Path(common.__file__).parent / 'hdl/fixed_pkg/'
        src = [common_path / 'fixed_pkg_c.vhd', common_path / 'fixed_float_types_c.vhd']
        src += self.get_conversion_sources()

        buffer = ""
        for key, value in rules.items():
            buffer += f"set_global_assignment -name {key} {value!s}\n"

        buffer += "\n"
        for file in src:
            buffer += f"set_global_assignment -name VHDL_FILE {file}\n"

        outpath = self.quartus_path / 'quartus_project.qsf'
        with outpath.open('w') as f:
            f.write(buffer)

        # this is just useless project file, enables opening from IDE
        outpath = self.quartus_path / 'quartus_project.qpf'
        with outpath.open('w') as f:
            f.write('PROJECT_REVISION = "quartus_project"')

    def make_quartus_netlist(self):
        self.logger.info('Running quartus map...will take time.')
        make_process = subprocess.call(['quartus_map', 'quartus_project'], cwd=self.quartus_path)
        assert make_process == 0

        self.logger.info('Running netlist writer.')
        make_process = subprocess.call(['quartus_eda', 'quartus_project'], cwd=self.quartus_path)
        assert make_process == 0

        return self.quartus_path / 'simulation/modelsim/quartus_project.vho'



import textwrap
from enum import Enum

from pyha.common.sfix import Sfix, ComplexSfix
from pyha.common.util import tabber
from pyha.conversion.converter import file_header
from pyha.conversion.coupling import pytype_to_vhdl, get_instance_vhdl_name


class NotTrainedError(Exception):
    pass


class NoInputsError(Exception):
    pass


class NoOutputsError(Exception):
    pass


class TopGenerator:
    def __init__(self, simulated_object):
        self.simulated_object = simulated_object

        # 0 or 1 calls wont propagate register outputs
        if self.simulated_object.main.calls == 0:
            raise NotTrainedError('Object must be trained > 1 times.')

        if len(self.get_object_inputs()) == 0:
            raise NoInputsError('Model has no inputs (arguments to main).')

        if len(self.get_object_return()) == 0:
            raise NoOutputsError('Model has no outputs (main returns).')

    def get_object_args(self) -> list:
        return list(self.simulated_object.main.last_args)

    def get_object_kwargs(self) -> list:
        return self.simulated_object.main.last_kwargs.items()

    def get_object_inputs(self) -> list:
        return self.get_object_args() + [x[1] for x in self.get_object_kwargs()]

    def get_object_return(self) -> list:
        rets = self.simulated_object.main.last_return
        if isinstance(rets, tuple):
            return list(rets)
        elif rets is None:
            return []
        else:
            return [rets]  # single value

    def pyvar_to_stdlogic(self, var) -> str:
        if type(var) == int:
            return 'std_logic_vector(31 downto 0)'
        elif type(var) == bool:
            return 'std_logic'
        elif type(var) in (Sfix, ComplexSfix):
            return var.to_stdlogic()
        elif isinstance(var, Enum):
            return self.pyvar_to_stdlogic(var.value)
        elif isinstance(var, list):
            if isinstance(var[0], bool):
                return f'std_logic_vector({len(var) - 1} downto 0)'
        else:
            assert 0

    def vhdl_slv_to_normal(self, var, var_name) -> str:
        if type(var) == int:
            return f'to_integer(signed({var_name}))'
        elif type(var) == bool:
            return f'logic_to_bool({var_name})'
        elif type(var) == Sfix:
            return f'Sfix({var_name}, {var.left}, {var.right})'
        elif type(var) == ComplexSfix:
            size = int(var.bitwidth())
            mid = size // 2
            real = f'Sfix({var_name}({size - 1} downto {mid}), {var.left}, {var.right})'
            imag = f'Sfix({var_name}({mid - 1} downto {0}), {var.left}, {var.right})'
            return f'(real=>{real}, imag=>{imag})'
        else:
            assert 0

    def normal_to_slv(self, var, var_name) -> str:
        if type(var) == int:
            return f'std_logic_vector(to_signed({var_name}, 32))'
        elif type(var) == bool:
            return f'bool_to_logic({var_name})'
        elif type(var) == Sfix:
            return f'to_slv({var_name})'
        elif type(var) == ComplexSfix:
            return f'to_slv({var_name}.real) & to_slv({var_name}.imag)'
        elif isinstance(var, Enum):
            return self.normal_to_slv(var.value, var_name)
        elif isinstance(var, list):
            if isinstance(var[0], bool):
                return f'bool_list_to_logic({var_name})'
        else:
            assert 0

    def make_entity_inputs(self) -> str:
        return '\n'.join(f'in{i}: in {self.pyvar_to_stdlogic(x)};'
                         for i, x in enumerate(self.get_object_inputs()))

    def make_entity_outputs(self) -> str:
        return '\n'.join(f'out{i}: out {self.pyvar_to_stdlogic(x)};'
                         for i, x in enumerate(self.get_object_return()))

    def make_output_variables(self) -> str:
        return '\n'.join(f'variable var_out{i}: {pytype_to_vhdl(x)};'
                         for i, x in enumerate(self.get_object_return()))

    def make_output_type_conversions(self) -> str:
        return '\n'.join(f'out{i} <= {self.normal_to_slv(x, "var_out{}".format(i))};'
                         for i, x in enumerate(self.get_object_return()))

    def make_input_variables(self) -> str:
        return '\n'.join(f'variable var_in{i}: {pytype_to_vhdl(x)};'
                         for i, x in enumerate(self.get_object_inputs()))

    def make_input_type_conversions(self) -> str:
        return '\n'.join(f'var_in{i} := {self.vhdl_slv_to_normal(x, "in{}".format(i))};'
                         for i, x in enumerate(self.get_object_inputs()))

    def make_complex_types(self):
        complex_vars = []
        for x in self.get_object_inputs() + self.get_object_return():
            if type(x) is ComplexSfix:
                new = x.vhdl_type_define()
                if new not in complex_vars:
                    complex_vars.append(new)
        return '\n'.join(x for x in complex_vars)

    def make_imports(self) -> str:
        return textwrap.dedent("""\
            library ieee;
                use ieee.std_logic_1164.all;
                use ieee.numeric_std.all;
                use ieee.fixed_pkg.all;
                use ieee.math_real.all;

            library work;
                use work.PyhaUtil.all;
                use work.ComplexTypes.all;
                use work.all;""")

    def object_class_name(self) -> str:
        # make sure we escape reserved names
        return get_instance_vhdl_name(self.simulated_object)

    def make_call_arguments(self) -> str:

        input_args = ', '.join(f'var_in{i}'
                               for i, _ in enumerate(self.get_object_args()))
        ofs = len(self.get_object_args())
        input_kwargs = ', '.join(f'{x[0]}=>var_in{i + ofs}'
                                 for i, x in enumerate(self.get_object_kwargs()))

        inputs = ', '.join([input_args, input_kwargs]) if len(self.get_object_kwargs()) else input_args

        outputs = ', '.join(f'ret_{i}=>var_out{i}'
                            for i, _ in enumerate(self.get_object_return()))

        return ', '.join([inputs, outputs])

    def make(self):
        template = textwrap.dedent("""\
                {FILE_HEADER}
                {IMPORTS}

                entity  top is
                    port (
                        clk, rst_n, enable: in std_logic;

                        -- inputs
                {ENTITY_INPUTS}

                        -- outputs
                {ENTITY_OUTPUTS}
                    );
                end entity;

                architecture arch of top is
                begin
                    process(clk, rst_n)
                        variable self: {DUT_NAME}.self_t;
                        -- input variables
                {INPUT_VARIABLES}

                        --output variables
                {OUTPUT_VARIABLES}
                    begin
                    if (not rst_n) then
                        {DUT_NAME}.\\_pyha_reset_self\\(self);
                    elsif rising_edge(clk) then
                        if enable then
                            --convert slv to normal types
                {INPUT_TYPE_CONVERSIONS}

                            --call the main entry
                            -- without this Quartus wont honor constants
                            {DUT_NAME}.\\_pyha_constants_self\\(self);
                            {DUT_NAME}.main(self, {CALL_ARGUMENTS});
                            {DUT_NAME}.\\_pyha_update_self\\(self);

                            --convert normal types to slv
                {OUTPUT_TYPE_CONVERSIONS}
                        end if;
                      end if;

                    end process;
                end architecture;""")

        def tab(x):
            return tabber(tabber(x))

        sockets = {}
        sockets['FILE_HEADER'] = file_header()
        sockets['DUT_NAME'] = self.object_class_name()
        sockets['IMPORTS'] = self.make_imports()
        sockets['ENTITY_INPUTS'] = tab(self.make_entity_inputs())
        sockets['ENTITY_OUTPUTS'] = tab(
            self.make_entity_outputs()[:-1])  # -1 removes the last ';', VHDL has some retarded rules
        sockets['INPUT_VARIABLES'] = tab(self.make_input_variables())
        sockets['OUTPUT_VARIABLES'] = tab(self.make_output_variables())
        sockets['INPUT_TYPE_CONVERSIONS'] = tabber(tab(self.make_input_type_conversions()))
        sockets['OUTPUT_TYPE_CONVERSIONS'] = tabber(tab(self.make_output_type_conversions()))
        sockets['CALL_ARGUMENTS'] = self.make_call_arguments()

        res = template.format(**sockets)

        return res



# Simple tests for an adder module

import cocotb
import numpy as np
from cocotb.clock import Clock
from cocotb.result import ReturnValue
from cocotb.triggers import RisingEdge, Timer, FallingEdge


@cocotb.coroutine
def reset(dut, duration=10000):
    dut.log.debug("Resetting DUT")
    dut.rst_n = 0
    yield Timer(duration)
    yield RisingEdge(dut.clk)
    dut.rst_n = 1
    dut.log.debug("Out of reset")


@cocotb.coroutine
def run_dut(dut, in_data, out_count):
    dut.enable = 1
    cocotb.fork(Clock(dut.clk, 5000).start())
    yield reset(dut)

    # FIXME: test input and output names and report error
    ret = []
    # print('Input data: {}'.format(in_data))
    for x in in_data:

        # put input
        # print('Processing slice: {}'.format(x))
        for i, xi in enumerate(x):
            # print('Set {} to {}'.format('in' + str(i), xi))
            setattr(dut, 'in' + str(i), int(xi.astype(int)))

        # NOTICE: need to have both yields to match simulation.
        yield RisingEdge(dut.clk)
        yield FallingEdge(dut.clk)

        # collect output
        tmp = []
        for i in range(out_count):
            var = 'out' + str(i)
            # val = getattr(dut, var).value.signed_integer
            val = str(getattr(dut, var).value)
            # print(val)
            tmp.append(val)
        ret.append(tmp)

    # print('Finish, ret: {}'.format(ret))
    raise ReturnValue(ret)


@cocotb.test()
def test_main(dut):
    import os
    in_data = np.load(os.getcwd() + '/../input.npy')
    # in_data = np.transpose(in_data)
    # print(in_data)

    output_vars = int(os.environ['OUTPUT_VARIABLES'])
    hdl_out = yield run_dut(dut, in_data, output_vars)
    # hdl_out = np.transpose(hdl_out)
    # print(hdl_out)

    np.save(os.getcwd() + '/../output.npy', hdl_out)



from pyha.common.hwsim import HW, PyhaFunc, SKIP_FUNCTIONS, is_convertible, PYHA_VARIABLES
from pyha.common.sfix import Sfix, ComplexSfix


class FunctionNotSimulated(Exception):
    def __init__(self, class_name, function_name):
        message = f'Function not simulated before conversion!\nClass: {class_name}\nFunction: {function_name}'
        super().__init__(message)


class VariableNotConvertible(Exception):
    def __init__(self, class_name, function_name, variable_name, variable):
        message = f'Variable not convertable!\nClass: {class_name}\nFunction: {function_name}\nVariable: {variable_name}\nValue: {type(variable)}:{variable}'
        super().__init__(message)


def extract_datamodel(obj):
    ret = {}
    for key, val in obj.__dict__['_pyha_initial_self'].__dict__.items():
        if key in PYHA_VARIABLES:
            continue
        if is_convertible(val):
            last = obj.__dict__[key]
            # for Sfix use the initial value but LATEST bounds
            if isinstance(val, Sfix):
                val = Sfix(val.init_val, last.left, last.right)
            elif isinstance(val, ComplexSfix):
                val = ComplexSfix(val.init_val, last.left, last.right)
            elif isinstance(val, list) and isinstance(val[0], Sfix):
                val = [Sfix(new_val.init_val, last_val.left, last_val.right) for new_val, last_val in zip(val, last)]
            elif isinstance(val, HW):
                val = last
            elif isinstance(val, list) and isinstance(val[0], HW):
                # list of submodules
                # set all to single id
                val = last
                first_id = val[0]._pyha_instance_id
                for x in val:
                    x._pyha_instance_id = first_id
            ret.update({key: val})
    return ret


def extract_locals(obj):
    ret = {}
    class_name = type(obj).__name__
    for method in dir(obj):
        if method in SKIP_FUNCTIONS:  continue
        call = getattr(obj, method)
        # if hasattr(call, 'knows_locals'):
        if isinstance(call, PyhaFunc):
            if call.calls == 0:
                raise FunctionNotSimulated(class_name, call.func.__name__)

            for key, val in call.locals.items():
                if not is_convertible(val):
                    raise VariableNotConvertible(class_name, call.func.__name__, key, val)

            ret[call.func.__name__] = call.locals

    return ret


class DataModel:
    def __init__(self, obj=None, self_data=None, locals=None):
        self.obj = obj
        if obj is None:
            self.self_data = None if self_data is None else self_data
            self.locals = None if locals is None else locals
            self.constants = {}
        else:
            dm = extract_datamodel(obj)

            constants = obj._pyha_constants

            # remove constants from self data
            dm_clean = {k: v for k, v in dm.items() if k not in constants}
            if len(dm_clean) == 0:
                dm_clean['much_dummy_very_wow'] = 0  # this simplifies many testing code
            loc = extract_locals(obj)
            self.self_data = dm_clean
            self.locals = loc
            self.constants = constants

    def __str__(self):
        return f'self_data: {self.self_data}\tlocals: {self.locals}'



import numpy as np

from pyha.common.const import Const
from pyha.common.hwsim import HW
from pyha.common.sfix import Sfix
from pyha.components.cordic import NCO


class FSKModulator(HW):
    # todo: this should handle 'sps', but needs raised clock rate then..
    def __init__(self, deviation, fs):
        self.fs = fs
        self.deviation = deviation

        self.sensitivity = 2 * np.pi * deviation / fs

        self.nco = NCO()

        # constants
        # / np.pi is to keep stuff in -1 to 1 range
        self.sensitivity_pos = Const(Sfix(self.sensitivity / np.pi, 0, -17))
        self.sensitivity_neg = Const(Sfix(-self.sensitivity / np.pi, 0, -17))
        self._delay = self.nco._delay

    def main(self, symbol):
        if symbol:
            phase_step = self.sensitivity_pos
        else:
            phase_step = self.sensitivity_neg

        cout = self.nco.main(phase_step)
        return cout

    def model_main(self, symbol_list):
        d_phase = 0
        phl = []
        for symbol in symbol_list:
            d_phase += self.sensitivity if symbol else -self.sensitivity  # this is FSK
            d_phase = ((d_phase + np.pi) % (2.0 * np.pi)) - np.pi  # keep in pi range
            phl.append(d_phase * 1j)

        return np.exp(phl)



from contextlib import suppress

import numpy as np

class SignalTapParser:
    def __init__(self, file: str):
        import csv

        self.labels = []
        self.data = []
        with open(file) as csvfile:
            reader = csv.reader(csvfile)
            for x in reader:
                if len(x) and x[0] == 'Data:':
                    self.labels = next(reader)
                    self.data = [x for x in reader]

        self.trans_data = np.array(self.data).T

    def __getitem__(self, item: str):
        i = self.labels.index(item)
        return self.trans_data[i]

    def to_int(self, data, bits):
        r = []
        for x in data:
            with suppress(ValueError):
                new = int(x, 16)
                if new >= 2**(bits-1): #conv to signed
                    new -= 2**(bits)
                r.append(new)
        return r

    # NB! these wont work if you export sfixed signal( has negative bounds in name )
    # need to invert msb or something
    def to_float(self, data, bits):
        """ assume 1 sign others fractional"""
        ints = self.to_int(data, bits)
        return np.array(ints) / 2 ** (bits-1)

    def to_bladerf(self, data):
        """ assume 5 bit is for integer(1 bit is sign) and others for fractional part
         This is used in bladeRF """
        ints = self.to_int(data, 16)
        return np.array(ints) / 2 ** (11)







import numpy as np

from pyha.common.const import Const
from pyha.common.hwsim import HW
from pyha.common.sfix import ComplexSfix, resize, Sfix, fixed_truncate
from pyha.components.cordic import Angle
from pyha.components.util_complex import Conjugate, ComplexMultiply


class QuadratureDemodulator(HW):
    """
    http://gnuradio.org/doc/doxygen-3.7/classgr_1_1analog_1_1quadrature__demod__cf.html#details

    :param gain: inverse of tx sensitivity. In RTL this is further multiplied by PI, because CORDIC returns angle in -1 to 1 range.
    """
    def __init__(self, gain):
        self.gain = gain

        # components / registers
        self.conjugate = Conjugate()
        self.complex_mult = ComplexMultiply()
        self.angle = Angle()
        self.out = Sfix()

        # constants
        # pi term puts angle output to pi range
        self.gain_sfix = Const(Sfix(self.gain * np.pi, 3, -14))

        self._delay = self.conjugate._delay + \
                     self.complex_mult._delay + \
                     self.angle._delay + 1

    def main(self, c):
        """

        :param c: baseband
        :type c: ComplexSfix
        :return: demodulated signal
        :rtype: Sfix
        """
        conj = self.conjugate.main(c)
        mult = self.complex_mult.main(c, conj)
        angle = self.angle.main(mult)
        fix_gain = resize(self.gain_sfix * angle, c.real, round_style=fixed_truncate)

        # output register
        self.next.out = fix_gain
        return self.out

    def model_main(self, c):
        """ Model that verification is ran against """
        demod = np.angle(c[1:] * np.conjugate(c[:-1]))
        fix_gain = self.gain * demod
        return fix_gain


# test modules
class QuadratureDemodulatorPartial0(HW):
    def __init__(self, gain):
        self.gain = gain
        self.conjugate = Conjugate()
        self.prev = ComplexSfix()
        self._delay = self.conjugate._delay

    def main(self, c):
        r = self.conjugate.main(c)
        return r

    def model_main(self, c):
        demod = np.conjugate(c[:-1])
        return demod


class QuadratureDemodulatorPartial1(HW):
    def __init__(self, gain):
        self.gain = gain

        self.conjugate = Conjugate()
        self.complex_mult = ComplexMultiply()
        self.prev = ComplexSfix()
        self._delay = self.conjugate._delay + \
                     self.complex_mult._delay

    def main(self, c):
        r = self.conjugate.main(c)
        ra = self.complex_mult.main(c, r)
        return ra

    def model_main(self, c):
        demod = c[1:] * np.conjugate(c[:-1])
        return demod



import logging
import os
from contextlib import suppress
from copy import deepcopy
from enum import Enum
from functools import wraps
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List

import numpy as np

from pyha.common.sfix import Sfix, ComplexSfix
from pyha.conftest import SKIP_SIMULATIONS_MASK
from pyha.simulation.sim_provider import SimProvider


class NoModelError(Exception):
    pass


class InputTypesError(Exception):
    pass


SIM_MODEL, SIM_HW_MODEL, SIM_RTL, SIM_GATE = ['MODEL', 'HW_MODEL', 'RTL', 'GATE']


def flush_pipeline(func):
    """ For inputs: adds 'x.get_delay()' dummy samples, to flush out pipeline values
    For outputs: removes the first 'x.get_delay()' samples, as these are initial pipeline values"""

    @wraps(func)
    def flush_pipeline_wrap(self, *args, **kwargs):
        delay = 0
        with suppress(AttributeError):  # no get_delay()
            delay = self.model._delay
        if delay == 0:
            return func(self, *args, **kwargs)

        args = list(args)

        for i in range(delay):
            args.append(args[0])

        ret = func(self, *args, **kwargs)
        ret = ret[delay:]
        return ret

    return flush_pipeline_wrap


def in_out_transpose(func):
    """ Transpose input before call and output after call """

    @wraps(func)
    def transposer_wrap(self, *args, **kwargs):
        # numpy cannot be used as it loses type info (converts everything to float)
        args = [x for x in zip(*args)]  # transpose

        ret = func(self, *args, **kwargs)

        with suppress(TypeError):  # was one dimensional list
            ret = [list(x) for x in zip(*ret)]  # transpose
        return ret

    return transposer_wrap


def type_conversions(func):
    @wraps(func)
    def type_enforcement_wrap(self, *args, **kwargs):
        # force input types
        if self.input_types is not None:
            args = [[to_type(x) for x in data] for data, to_type in zip(args, self.input_types)]

        ret = func(self, *args, **kwargs)

        def output_types(li):
            ret = []
            for x in li:
                if type(x) in [list, tuple]:
                    ret.append(output_types(x))
                elif type(x) == Sfix:
                    ret.append(float(x))
                elif type(x) == ComplexSfix:
                    ret.append(float(x.real) + float(x.imag) * 1j)
                elif isinstance(x, Enum):
                    ret.append(x.value)
                else:
                    ret.append(x)
            return ret

        ret = output_types(ret)
        return np.array(ret)

    return type_enforcement_wrap


class Simulation:
    """ Returned stuff is always Numpy array? """
    hw_instances = {}

    def __init__(self, simulation_type, model=None, input_types: List[object] = None, dir_path=None):
        self.logger = logging.getLogger(__name__)
        # self.tmpdir = TemporaryDirectory().name
        self.dir_path = dir_path
        if self.dir_path is None:
            self.keep_me_alive = TemporaryDirectory()
            self.dir_path = self.keep_me_alive.name
        else:
            self.dir_path = str(Path(self.dir_path).expanduser())
            if not Path(self.dir_path).exists():
                os.makedirs(self.dir_path)

        self.input_types = []
        self.model = None
        self.sim = None
        self.cocosim = None
        self.simulation_type = simulation_type

        # direct output from dut call will be written here( without type conversions, pipeline fixes..)
        self.pure_output = []
        # if model is not None:
        self.init(model, input_types)

    def init(self, model=None, input_types: List[object] = None):
        self.model = model
        self.input_types = input_types
        if self.model is None:
            raise NoModelError('Trying to run simulation but "model" is None')

        if not hasattr(self.model, 'main') and self.simulation_type in (SIM_HW_MODEL, SIM_RTL, SIM_GATE):
            raise NoModelError('Your model has no "main" function')

        if not hasattr(self.model, 'model_main') and self.simulation_type is SIM_MODEL:
            raise NoModelError('Trying to run "SIM_MODEL" simulation but your model has no "model_main" function!')

        # save ht HW model for conversion
        if self.simulation_type == SIM_HW_MODEL:
            Simulation.hw_instances[model.__class__.__name__] = model

    def prepare_hw_simulation(self):
        # grab the already simulated model!
        self.model = Simulation.hw_instances[self.model.__class__.__name__]
        self.sim = SimProvider(Path(self.dir_path), self.model, self.simulation_type)
        return self.sim.main()
        # conv = Conversion(self.model)
        # quartus = None
        # if self.simulation_type is SIM_GATE:
        #     make_quartus_project(Path(self.tmpdir.name), conv)
        # return CocotbAuto(Path(self.tmpdir.name), conv)

    @type_conversions
    @in_out_transpose
    @flush_pipeline
    def hw_simulation(self, *args):
        if self.simulation_type == SIM_HW_MODEL:
            # reset registers, in order to match COCOTB RTL simulation behaviour
            self.model.__dict__.update(deepcopy(self.model._pyha_initial_self).__dict__)
            ret = []
            for x in args:
                ret.append(self.model.main(*x))
                self.model._pyha_update_self()
        elif self.simulation_type in [SIM_RTL, SIM_GATE]:
            ret = self.cocosim.run(*args)
        else:
            assert 0

        self.pure_output = ret
        return ret

    def main(self, *args) -> np.array:
        self.logger.info(f'Running {self.simulation_type} simulation!')
        # test if user provided legal 'input_types'
        if self.simulation_type is not SIM_MODEL or self.input_types is not None:  # it is legal to not pass input_types if SIM_MODEL
            if self.input_types is None or (len(args) != len(self.input_types)):
                raise InputTypesError('Your "Simulation(input_types=X)" does not match arguements to "main" function!')

        # test that there are no Sfix arguments
        for x in args:
            if type(x) is list:
                if type(x[0]) is Sfix:
                    raise InputTypesError(
                        'You are passing Sfix values to your model, pass float instead (will be converted to sfix automatically)!')

        if self.simulation_type in (SIM_RTL, SIM_GATE) and self.cocosim is None:
            self.cocosim = self.prepare_hw_simulation()

        if self.simulation_type == SIM_MODEL:
            r = self.model.model_main(*args)

            if r == []:
                return np.array(r)

            if isinstance(r, tuple):
                if isinstance(r[0], list):
                    # assume that no transpose needed ( model returns correct way )
                    return np.array(r)
                return np.transpose(r)

            if isinstance(r, list) and isinstance(r[0], tuple):
                return np.transpose(r)
            return np.array(r)
        else:
            return self.hw_simulation(*args)


##############################################################################
##############################################################################
##############################################################################
##############################################################################
##############################################################################
##############################################################################
##############################################################################
##############################################################################
##############################################################################
# utility functions

def debug_assert_sim_match(model, types, expected, *x, simulations=None, rtol=1e-05, atol=1e-9, dir_path=None,
                           fuck_it=False, **kwards):
    """ Instead of asserting anything return outputs of each simulation """
    simulations = sim_rules(simulations, model)
    outs = []
    for sim_type in simulations:
        dut = Simulation(sim_type, model=model, input_types=types, dir_path=dir_path)
        hw_y = dut.main(*x)
        outs.append(hw_y)
    return outs


def assert_model_hwmodel_match(model, types, *x, rtol=1e-9, atol=1e-9):
    if skipping_hwmodel_simulations() or skipping_model_simulations():
        return
    outs = debug_assert_sim_match(model, types, [1], *x, simulations=[SIM_MODEL, SIM_HW_MODEL])
    # return outs
    np.testing.assert_allclose(outs[0], outs[1], rtol, atol)


def assert_hwmodel_rtl_match(model, types, *x):
    if skipping_hwmodel_simulations() or skipping_rtl_simulations():
        return
    outs = debug_assert_sim_match(model, types, [1], *x, simulations=[SIM_HW_MODEL, SIM_RTL])
    np.testing.assert_allclose(outs[0], outs[1], rtol=1e-9)


def plot_assert_sim_match(model, types, expected, *x, simulations=None, rtol=1e-05, atol=1e-9, dir_path=None, skip_first=0):
    """
    Same arguments as :code:`assert_sim_match`. Instead of asserting it plots all the simulations.

    """
    import matplotlib.pyplot as plt
    simulations = sim_rules(simulations, model)
    for sim_type in simulations:
        dut = Simulation(sim_type, model=model, input_types=types, dir_path=dir_path)
        hw_y = dut.main(*x)
        plt.plot(hw_y[skip_first:], label=str(sim_type))

    plt.legend()
    plt.show()


def assert_sim_match(model, types, expected, *x, simulations=None, rtol=1e-05, atol=1e-9, dir_path=None, skip_first=0):
    """
    Run bunch of simulations and assert that they match outputs.

    :param model: Instance of your class
    :param types: If :code:`main` is defined, provide input types for each argument, all arguments will be
     automatically casted to these types.
    :param expected: Expected output of the simulation. If None, assert all simulations match eachother.
    :param x: Inputs, if you have multiple inputs, use *x for unpacking.
    :param simulations: Simulations to run and assert:
    - SIM_MODEL: runs model ('model_main')
    - SIM_HW_MODEL: runs HW model ('main')
    - SIM_RTL: converts to VHDL and runs RTL simulation via GHDL and Cocotb
    - SIM_GATE: runs sources trough Quartus and simulates the generated netlist
    .. note:: If None(default), runs all simulations. SIM_HW_MODEL must be run if SIM_RTL or SIM_GATE are going to run.
    :param rtol: Relative tolerance for assertion. Look np.testing.assert_allclose.
    :param atol: Absolute tolerance for assertion. Look np.testing.assert_allclose.
    :param dir_path: Where are conversion outputs written, if empty uses temporary directory.
    :param skip_first: Skip first 'n' samples for assertion.

    """
    l = logging.getLogger(__name__)
    simulations = sim_rules(simulations, model)

    for sim_type in simulations:
        dut = Simulation(sim_type, model=model, input_types=types, dir_path=dir_path)
        hw_y = dut.main(*x)
        if expected is None and sim_type is simulations[0]:
            expected = hw_y

        try:
            assert len(expected) > 0
            # if type(expected[0]) != type(hw_y[0]):
            #     hw_y = hw_y.astype(type(expected[0]))
            np.testing.assert_allclose(expected[skip_first:], hw_y[skip_first:len(expected)], rtol, atol=atol)
        except AssertionError as e:
            l.error('##############################################################')
            l.error('##############################################################')
            l.error(f'\t\t"{sim_type}" failed')
            l.error('##############################################################')
            l.error('##############################################################')

            raise


def skipping_gate_simulations():
    if SKIP_SIMULATIONS_MASK & 8:
        return True
    with suppress(KeyError):
        if int(os.environ['PYHA_SKIP_QUARTUS_SIMS']):
            return True
    return False


def skipping_rtl_simulations():
    return SKIP_SIMULATIONS_MASK & 4


def skipping_hwmodel_simulations():
    return SKIP_SIMULATIONS_MASK & 2


def skipping_model_simulations():
    return SKIP_SIMULATIONS_MASK & 1


def sim_rules(simulations, model):
    if simulations is None:
        simulations = [SIM_MODEL, SIM_HW_MODEL, SIM_RTL, SIM_GATE]
    # force simulation rules, for example SIM_RTL cannot be run without SIM_HW_MODEL, that needs to be ran first.
    assert simulations in [[SIM_MODEL],
                           [SIM_MODEL, SIM_HW_MODEL],
                           [SIM_MODEL, SIM_HW_MODEL, SIM_RTL],
                           [SIM_MODEL, SIM_HW_MODEL, SIM_GATE],
                           [SIM_HW_MODEL],
                           [SIM_HW_MODEL, SIM_RTL],
                           [SIM_MODEL, SIM_HW_MODEL, SIM_RTL, SIM_GATE],
                           [SIM_HW_MODEL, SIM_RTL, SIM_GATE],
                           [SIM_HW_MODEL, SIM_GATE]]

    if not hasattr(model, 'model_main') and SIM_MODEL in simulations:
        simulations.remove(SIM_MODEL)
        logging.getLogger(__name__).warning('Skipping MODEL simulation, because there is no "model_main" function!')


    # for travis build, skip all the tests involving quartus
    if skipping_model_simulations() and SIM_MODEL in simulations:
        simulations.remove(SIM_MODEL)
        logging.getLogger(__name__).warning('########## SKIPPING MODEL SIMULATIONS ##########')

    if skipping_hwmodel_simulations() and SIM_HW_MODEL in simulations:
        simulations.remove(SIM_HW_MODEL)
        logging.getLogger(__name__).warning('########## SKIPPING HW_MODEL SIMULATIONS ##########')

    if skipping_rtl_simulations() and SIM_RTL in simulations:
        simulations.remove(SIM_RTL)
        logging.getLogger(__name__).warning('########## SKIPPING RTL SIMULATIONS ##########')

    if skipping_gate_simulations() and SIM_GATE in simulations:
        simulations.remove(SIM_GATE)
        logging.getLogger(__name__).warning('########## SKIPPING GATE SIMULATIONS ##########')

    return simulations



from pyha.common.hwsim import HW
# from pyha.common.sfix import resize
from pyha.common.sfix import resize, Sfix
from pyha.components.moving_average import MovingAverage


class DCRemoval(HW):
    def __init__(self, window_len):
        self.mavg = [MovingAverage(window_len) for _ in range(2)]
        self.delay_x = Sfix()
        self.delay_x2 = Sfix()
        self.out = Sfix()

        self._delay = 3

    def main(self, x):

        # run sample trough all moving averages
        tmp = x
        for mav in self.mavg:
            tmp = mav.main(tmp)

        self.next.delay_x = x
        self.next.delay_x2 = self.delay_x
        self.next.out = resize(self.delay_x2 - tmp, size_res=x)
        return self.out

    def model_main(self, x):
        tmp = x
        for mav in self.mavg:
            tmp = mav.model_main(tmp)

        return x - tmp



from pyha.common.hwsim import HW
from pyha.common.sfix import ComplexSfix
from pyha.components.blade_demod.blade_adaptor import NormalToBlade, BladeToComplex
from pyha.components.moving_average import MovingAverage
from pyha.components.quadrature_demodulator import QuadratureDemodulator


class BladeDemodPartial1(HW):
    def __init__(self, quadrature_demodulator_gain, moving_average_length):
        self.blade_to_complex = BladeToComplex()
        self.quadrature_demodulator = QuadratureDemodulator(quadrature_demodulator_gain)
        self.moving_average = MovingAverage(moving_average_length)

        self.normal_to_blade = NormalToBlade()
        # self._delay = self.quadrature_demodulator.get_delay() + self.blade_to_complex.get_delay()\
        #               + self.moving_average.get_delay()\
        #               + self.normal_to_blade.get_delay()

        self._delay = self.quadrature_demodulator._delay\
                      + self.moving_average._delay

    def main(self, i, q):
        # c = self.next.blade_to_complex.main(i, q)
        c = ComplexSfix(i, q)
        demod = self.quadrature_demodulator.main(c)
        mavg = self.moving_average.main(demod)
        # blade = self.next.normal_to_blade.main(demod)
        return mavg

    def model_main(self, i, q):
        # c = self.blade_to_complex.model_main(i, q)
        c = i + q * 1j
        a = self.quadrature_demodulator.model_main(c)
        b = self.moving_average.model_main(a)
        return b
            # return self.normal_to_blade.model_main(b)


class DemodQuadMavg(HW):
    def __init__(self, quadrature_demodulator_gain, moving_average_length):
        self.quadrature_demodulator = QuadratureDemodulator(quadrature_demodulator_gain)
        self.moving_average = MovingAverage(moving_average_length)

        self._delay = self.quadrature_demodulator._delay \
                      + self.moving_average._delay

    def main(self, c):
        demod = self.quadrature_demodulator.main(c)
        mavg = self.moving_average.main(demod)
        return mavg

    def model_main(self, c):
        a = self.quadrature_demodulator.model_main(c)
        b = self.moving_average.model_main(a)
        return b


class DemodQuad(HW):
    def __init__(self, quadrature_demodulator_gain):
        self.quadrature_demodulator = QuadratureDemodulator(quadrature_demodulator_gain)
        self._delay = self.quadrature_demodulator._delay

    def main(self, c):
        demod = self.quadrature_demodulator.main(c)
        return demod

    def model_main(self, c):
        a = self.quadrature_demodulator.model_main(c)
        return a

class BladeDemodQuad(HW):
    def __init__(self, quadrature_demodulator_gain):
        self.quadrature_demodulator = QuadratureDemodulator(quadrature_demodulator_gain)
        self._delay = self.quadrature_demodulator._delay

    def main(self, c):
        demod = self.quadrature_demodulator.main(c)
        return demod

    def model_main(self, c):
        a = self.quadrature_demodulator.model_main(c)
        return a


# for testing
class BladeDemodPartial0(HW):
    def __init__(self):
        self.blade_to_complex = BladeToComplex()
        self.quadrature_demodulator = QuadratureDemodulator(0.5)
        self.normal_to_blade = NormalToBlade()
        self._delay = self.quadrature_demodulator._delay + self.blade_to_complex._delay\
                      + self.normal_to_blade._delay

    def main(self, i, q):
        c = self.blade_to_complex.main(i, q)
        demod = self.quadrature_demodulator.main(c)
        blade = self.normal_to_blade.main(demod)
        return blade

    def model_main(self, i, q):
        c = self.blade_to_complex.model_main(i, q)
        a = self.quadrature_demodulator.model_main(c)
        return self.normal_to_blade.model_main(a)





import sys
from copy import deepcopy, copy
from enum import Enum

import numpy as np
from six import iteritems, with_metaclass

from pyha.common.const import Const
from pyha.common.sfix import Sfix, ComplexSfix

"""
Purpose: Make python class simulatable as hardware, mainly provide 'register' behaviour
"""

# functions that will not be decorated/converted/parsed
SKIP_FUNCTIONS = ('__init__', 'model_main')

# Pyha related variables in the object __dict__
PYHA_VARIABLES = ('_pyha_constants', '_pyha_initial_self', 'next', '_pyha_submodules', '_pyha_instance_id', '_delay')


class AssignToSelf(Exception):
    def __init__(self, class_name, variable_name):
        message = f'Assigment to self.{variable_name}, did you mean self.next.{variable_name}?\nClass: {class_name}'
        super().__init__(message)


class TypeNotConsistent(Exception):
    def __init__(self, class_name, function_name, variable_name, old, new):
        # these clutter printing
        from contextlib import suppress
        with suppress(KeyError):  # only available for 'self'
            new.pop('_pyha_initial_self')
            old.pop('_pyha_initial_self')
        message = f'Self/local not consistent type!\nClass: {class_name}\nFunction: {function_name}\nVariable: {variable_name}\nOld: {type(old)}:{repr(old)}\nNew: {type(new)}:{new}'
        super().__init__(message)


def is_convertible(obj):
    allowed_types = [ComplexSfix, Sfix, int, bool, Const]
    if type(obj) in allowed_types:
        return True
    elif isinstance(obj, list):
        # To check whether all elements are of the same type
        if len(set(map(type, obj))) == 1:
            if all(type(x) in allowed_types for x in obj):
                return True
            elif isinstance(obj[0], HW):  # list of submodules
                return True
    elif isinstance(obj, Enum):
        return True
    elif isinstance(obj, HW):
        return True

    return False

def deepish_copy(org):
    """
    https://writeonly.wordpress.com/2009/05/07/deepcopy-is-a-pig-for-simple-data/
    much, much faster than deepcopy, for a dict of the simple python types.
    """
    out = dict().fromkeys(org)
    for k, v in iteritems(org):
        try:
            out[k] = v.copy()  # dicts, sets
        except AttributeError:
            try:
                out[k] = v[:]  # lists, tuples, strings, unicode
            except TypeError:
                # Without this assign to imag or real will fuck up everything
                out[k] = deepcopy(v)
                # if isinstance(v, ComplexSfix):
                #     out[k] = copy(v)
                # else:
                #     out[k] = v  # ints

    return out


class PyhaFunc:
    class TraceManager:
        """ Enables nested functions calls, thanks to ref counting """
        last_call_locals = {}
        refcount = 0

        @classmethod
        def tracer(cls, frame, event, arg):
            # Note: this runs for ALL returns, only the LAST frame is valid info
            if event == 'return':
                cls.last_call_locals = frame.f_locals.copy()

        @classmethod
        def set_profile(cls):
            cls.refcount += 1
            sys.setprofile(cls.tracer)

        @classmethod
        def remove_profile(cls):
            cls.refcount -= 1
            assert cls.refcount >= 0
            sys.setprofile(None)

        @classmethod
        def restore_profile(cls):
            if cls.refcount > 0:
                sys.setprofile(cls.tracer)

    def __init__(self, func):
        self.class_name = func.__self__.__class__.__name__
        self.function_name = func.__name__
        self.func = func
        self.calls = 0
        self.locals = {}

        # used for top_generator
        self.last_args = {}
        self.last_kwargs = {}
        self.last_return = {}

        self.is_main = self.function_name == 'main'

    def dict_types_consistent_check(self, new, old):
        """ Check 'old' dict against 'new' dict for types, if not consistent raise """
        for key, value in new.items():
            if key in old:
                old_value = old[key]
                if isinstance(value, (Sfix, ComplexSfix)):
                    if value.left != old_value.left or value.right != old_value.right:
                        if old_value.left == 0 and old_value.right == 0:
                            # sfix lazy init
                            continue
                        elif value.right == 0 and value.left == 0:
                            # sfix lazy init, can happen for pipelines
                            continue
                        elif value.val == old_value.init_val:
                            # this is a shady condition, it helps against sfix values propagating trough pipelines, but may also mask valid errors
                            # HERE is reason why sometimes consistency check fails!
                            continue
                        elif old_value.val == old_value.init_val:
                            # shady shady stuff, helps if initival junk value is resized..
                            continue

                        raise TypeNotConsistent(self.class_name, self.function_name, key, old, new)
                elif type(value) != type(old_value):
                    raise TypeNotConsistent(self.class_name, self.function_name, key, old, new)
                elif isinstance(value, list):
                    if len(old_value) != len(value):
                        raise TypeNotConsistent(self.class_name, self.function_name, key, old, new)

    def forbid_assign_to_self(self, new, old):
        """ User should only assign to self.next.X, any assign to
            'self.X' is a bug and this decorator tests for that """

        for key, value in new.items():
            if key == 'next' or isinstance(value, (np.ndarray, np.generic)):
                continue

            if value != old[key]:
                raise AssignToSelf(self.class_name, key)

    def call_with_locals_discovery(self, *args, **kwargs):
        """ Call decorated function with tracing to read back local values """
        self.TraceManager.set_profile()
        res = self.func(*args, **kwargs)
        self.TraceManager.remove_profile()

        self.TraceManager.last_call_locals.pop('self')
        self.dict_types_consistent_check(self.TraceManager.last_call_locals, self.locals)

        self.locals.update(self.TraceManager.last_call_locals)

        # in case nested call, restore the tracer function
        self.TraceManager.restore_profile()
        return res

    def __call__(self, *args, **kwargs):

        self.last_args = args
        self.last_kwargs = kwargs
        real_self = self.func.__self__
        self.calls += 1

        # # CALL IS HERE!
        ret = self.call_with_locals_discovery(*args, **kwargs)

        self.last_return = ret

        real_self._outputs.append(ret)
        return ret


class Meta(type):
    """
    https://blog.ionelmc.ro/2015/02/09/understanding-python-metaclasses/#python-2-metaclass
    """
    instance_count = 0

    def validate_datamodel(cls, dict):
        # if list of submodules, make sure all 'constants' are the same
        for x in dict.values():
            if isinstance(x, list) and isinstance(x[0], HW):
                ref = x[0]._pyha_constants
                for listi in x:
                    di = listi._pyha_constants
                    if di != ref:
                        raise Exception(
                            f'List of submodules: {x}\n but constants are not equal!\n\nTry to remove Const() keyword.')

    def handle_constants(cls, dict):
        """ Go over dict and find all the constants. Remove the Const() wrapper
        and insert to _pyha_constants."""

        dict['_pyha_constants'] = {}
        for k, v in dict.items():
            if isinstance(v, Const):
                dict['_pyha_constants'][k] = v.value
                dict[k] = v.value

        # turn '_delay' into constant
        if '_delay' in dict:
            dict['_pyha_constants']['_delay'] = dict['_delay']

        return dict

    # ran when instance is made
    def __call__(cls, *args, **kwargs):
        ret = super(Meta, cls).__call__(*args, **kwargs)
        ret.__dict__ = cls.handle_constants(ret.__dict__)
        cls.validate_datamodel(ret.__dict__)

        ret._pyha_instance_id = cls.instance_count
        cls.instance_count += 1


        # give self.next to the new object
        ret.__dict__['next'] = type('next', (object,), {})()

        for k, v in ret.__dict__.items():
            if isinstance(v, HW) \
                    or k in PYHA_VARIABLES \
                    or (isinstance(v, list) and isinstance(v[0], HW)):
                continue
            if is_convertible(v):
                setattr(ret.next, k, deepcopy(v))

        # registery of submodules that need 'self update'
        ret._pyha_submodules = []
        for k, v in ret.__dict__.items():
            if k in PYHA_VARIABLES:
                continue
            if isinstance(v, HW):
                ret._pyha_submodules.append(v)
            elif isinstance(v, list) and v != [] and isinstance(v[0], HW):
                ret._pyha_submodules.append(v)

        # save the initial self values
        # all registers will be derived from these values!
        ret.__dict__['_pyha_initial_self'] = deepcopy(ret)

        # every call to 'main' will append returned values here
        ret._outputs = []

        # decorate all methods
        for method_str in dir(ret):
            if method_str in SKIP_FUNCTIONS:
                continue
            method = getattr(ret, method_str)
            if method_str[:2] != '__' and method_str[:1] != '_' and callable(
                    method) and method.__class__.__name__ == 'method':
                new = PyhaFunc(method)
                setattr(ret, method_str, new)

        return ret

class HW(with_metaclass(Meta)):
    """ For metaclass inheritance """

    def __deepcopy__(self, memo):
        """ http://stackoverflow.com/questions/1500718/what-is-the-right-way-to-override-the-copy-deepcopy-operations-on-an-object-in-p """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            # todo: maybe this also works for 'next'
            if k == '_pyha_initial_self':  # dont waste time on endless deepcopy
                setattr(result, k, copy(v))
            else:
                setattr(result, k, deepcopy(v, memo))
        return result

    def _pyha_update_self(self):
        self.__dict__.update(deepish_copy(self.next.__dict__))

        # update submodules
        for x in self._pyha_submodules:
            if isinstance(x, list):
                for item in x:
                    item._pyha_update_self()
            else:
                x._pyha_update_self()



from enum import Enum
from math import ceil

from pyha.common.const import Const
from pyha.common.hwsim import HW
from pyha.common.sfix import Sfix
from pyha.common.util import hex_to_bool_list


class DecodeState(Enum):
    HIGH_IN, HIGH_OUT, LOW_IN, LOW_OUT = range(4)


class BitsDecode(HW):
    # debugi = []
    # debugb = []
    # debugx = []
    # debugs = []
    # di = 0
    def __init__(self, decision_lim=0.2):
        self.decision_lim = Const(Sfix(decision_lim, 0, -17))
        self.bit_counter = 0
        self.state = False
        self.cstate = False
        self.period = Const(16)

        self.out_bit = False
        self.out_valid = False

        self.state = DecodeState.HIGH_IN
        self.sample_clock = 1

        self._delay = 1

    def main(self, x):
        self.next.out_valid = False
        self.next.sample_clock = self.sample_clock - 1

        # state machine -> ghetto clock recovery
        if self.state == DecodeState.HIGH_IN:
            if x < self.decision_lim:
                self.next.state = DecodeState.HIGH_OUT
                self.next.sample_clock = self.period
        elif self.state == DecodeState.HIGH_OUT:
            if x < -self.decision_lim:
                self.next.state = DecodeState.LOW_IN
                self.next.sample_clock = 0
        elif self.state == DecodeState.LOW_OUT:
            if x > self.decision_lim:
                self.next.state = DecodeState.HIGH_IN
                self.next.sample_clock = 0
        elif self.state == DecodeState.LOW_IN:
            if x > -self.decision_lim:
                self.next.state = DecodeState.LOW_OUT
                self.next.sample_clock = self.period

        # bit sampler
        if self.next.sample_clock == 0:
            if self.state == DecodeState.HIGH_IN:
                self.next.out_bit = True
            elif self.state == DecodeState.HIGH_OUT:
                self.next.out_bit = False
            elif self.state == DecodeState.LOW_IN:
                self.next.out_bit = False
            elif self.state == DecodeState.LOW_OUT:
                self.next.out_bit = True
            self.next.out_valid = True
            self.next.sample_clock = self.period

        # self.di += 1
        # self.debugx.append(x)
        # self.debugs.append(self.state.value/3)
        # if self.next.out_valid:
        #     self.debugi.append(self.di)
        #     self.debugb.append(self.decision_lim if self.next.out_bit else -self.decision_lim)
        return self.out_bit, self.out_valid

    def model_main(self, sig):
        # model will not match in case of noise signals
        # NB! likely that the hw algorithm actually works better than this model
        state = 0
        bit_counter = 0
        bits = []
        debugi = []
        debugb = []

        def push_bit(val, i):
            bits.append(val)
            debugi.append(i)
            debugb.append(self.decision_lim if val else -self.decision_lim)

        for i, x in enumerate(sig):
            bit_counter += 1
            if x > self.decision_lim:
                if state != 1:
                    push_bit(1, i)
                    bit_counter = 0
                state = 1
            elif x < -self.decision_lim:
                if state != 0:
                    push_bit(0, i)
                bit_counter = 0
                state = 0
            elif bit_counter >= self.period:
                n = 1 if state == 0 else 0
                push_bit(n, i)
                bit_counter = 0
        # #
        # import matplotlib.pyplot as plt
        # plt.plot(sig)
        # plt.stem(debugi, debugb)
        # plt.show()
        return bits


class CRC16(HW):
    def __init__(self, init_galois, xor):
        self.xor = Const(xor)
        # NB! tools generally raport fibo init value...need to convert it!
        self.init_galois = Const(init_galois)
        self.lfsr = init_galois

        self._delay = 1

    def main(self, din, reload):
        if reload:
            lfsr = self.init_galois
        else:
            lfsr = self.lfsr
        out = lfsr & 0x8000
        self.next.lfsr = ((lfsr << 1) | din) & 0xFFFF
        if out != 0:
            self.next.lfsr = self.next.lfsr ^ self.xor
        return self.lfsr

    def model_main(self, data, reload):
        ret = []
        lfsr = self.init_galois
        for din, rl in zip(data, reload):
            if rl:
                lfsr = self.init_galois

            out = lfsr & 0x8000
            lfsr = ((lfsr << 1) | din) & 0xFFFF
            if out:
                lfsr ^= self.xor
            ret.append(lfsr)
        return ret


# noinspection PyAugmentAssignment
class HeaderCorrelator(HW):
    def __init__(self, header, packet_len):
        # once header is found, 'packet_len' bits are skipped before next header can be correlated!
        # NB/TODO: there is possibility that random noise triggers the cooldown, so correct package may be lost!
        # TODO: why no output register?
        self.cooldown_reset = Const(packet_len - 1)
        self.header = Const(hex_to_bool_list(header))

        self.cooldown = 0
        self.shr = [False] * 16

        self._delay = 16

    # noinspection PyAugmentAssignment
    def main(self, din):
        self.next.shr = self.shr[1:] + [din]
        ret = False
        # noinspection PyAugmentAssignment
        if self.cooldown == 0:
            if self.shr == self.header:
                self.next.cooldown = self.cooldown_reset
                ret = True
        else:
            self.next.cooldown = self.next.cooldown - 1
        return ret

    def model_main(self, data):
        rets = [False] * len(data)
        i = 0
        while i < len(data):
            word = data[i:i + 16]
            if word == self.header:
                rets[i] = True
                i += self.cooldown_reset
            i += 1
        return rets


def bits_to_int(bits):
    assert len(bits) <= 32
    s = ''.join(str(int(x)) for x in bits)
    return int(s, 2)


# noinspection PyAugmentAssignment
class PacketSync(HW):
    def __init__(self, header, packet_len):
        self.packet_len_lim = Const(packet_len - 1)
        self.n32out = ceil(self.packet_len_lim.value / 32)  # number of 32 bit output packets per correct packet

        self.headercorr = HeaderCorrelator(header, packet_len)
        self.crc = CRC16(0x48f9, 0x1021)
        self.delay = [False] * self.headercorr._delay

        self.packet_counter = 0

        # read this when valid=1
        self.bits = [False] * (packet_len + self.headercorr._delay + self.crc._delay)

        self.part_out_counter = self.n32out
        self.armed = False
        # output
        self.out = [False] * 32
        self.valid = False

        self._delay = self.headercorr._delay + self.crc._delay + 1 + self.n32out  # n32out: delay is needed if outputs are to be presented!

        # constants
        self.n32out = Const(self.n32out)

    # noinspection PyAugmentAssignment
    def main(self, data):
        self.next.valid = False
        self.next.bits = self.bits[1:] + [data]

        reload = self.headercorr.main(data)
        self.next.packet_counter = self.packet_counter - 1
        if reload or self.packet_counter == 0:
            self.next.packet_counter = self.packet_len_lim
            self.next.armed = True

        self.next.delay = self.delay[1:] + [data]
        crc = self.crc.main(self.delay[0], reload)

        if self.armed and crc == 0 and self.packet_counter == 0:
            self.next.part_out_counter = 0
            if not reload:  # this may happen in case of two back to back packets!
                self.next.armed = False  # protect of fake packages

        if self.part_out_counter < self.n32out:
            self.next.part_out_counter = self.next.part_out_counter + 1

        for i in range(self.n32out):
            if self.part_out_counter == i:
                self.next.out = self.bits[(i * 32) - i: 32 + (i * 32) - i]
                self.next.valid = True

        return self.out, self.valid

    def model_main(self, data):
        head = self.headercorr.model_main(data)
        crc = self.crc.model_main(data, head)

        ret = []
        for i in range(self.packet_len_lim, len(crc)):
            b = i - self.packet_len_lim
            if head[b] and crc[i] == 0:
                for i in range(0, self.n32out):
                    ofs = i * 32
                    r = data[b + ofs: b + ofs + 32]
                    ret.append(r)
        return ret


class DemodToPacket(HW):
    def __init__(self):
        self.bits = BitsDecode(0.2)
        self.packsync = PacketSync(header=0x8dfc, packet_len=12 * 16)

        self.default_ret = Const([False] * 32)

        self._delay = self.bits._delay + self.packsync._delay

    def main(self, inp):
        pack_part = self.default_ret
        part_valid = False
        bit, bit_valid = self.bits.main(inp)
        if bit_valid:
            pack_part, part_valid = self.packsync.main(bit)
        return pack_part, part_valid

    def model_main(self, inp):
        bits = self.bits.model_main(inp)
        pack = self.packsync.model_main(bits)
        return pack



import numpy as np

from pyha.common.const import Const
from pyha.common.hwsim import HW
from pyha.common.sfix import Sfix, left_index, right_index, fixed_wrap
from pyha.common.sfix import resize
from pyha.common.util import is_power2


class MovingAverage(HW):
    """
    Moving average algorithm.

    :param window_len: Size of the moving average window, must be power of 2
    """

    def __init__(self, window_len):
        if window_len < 2:
            raise AttributeError('Window length must be >= 2')

        if not is_power2(window_len):
            raise AttributeError('Window length must be power of 2')

        self.window_len = window_len

        # registers
        # Sfix() -> bounds will be derived from simulation
        self.shift_register = [Sfix()] * self.window_len
        self.sum = Sfix()

        # constants
        self.window_pow = Const(int(np.log2(window_len)))

        # module delay
        self._delay = 1

    def main(self, x):
        """
        This works by keeping a history of 'window_len' elements and sum of them.
        Every clock last element will be subtracted and new added to the sum.
        Sum is then divided by the 'window_len'.
        More good infos: https://www.dsprelated.com/showarticle/58.php

        :param x: input to average
        :return: averaged output
        :rtype: Sfix
        """

        # add new element to shift register
        self.next.shift_register = [x] + self.shift_register[:-1]

        # calculate new sum
        nsum = self.sum + x - self.shift_register[-1]

        # resize sum, overflow is impossible
        self.next.sum = resize(nsum,
                               left_index=self.window_pow + left_index(x),
                               right_index=right_index(x),
                               overflow_style=fixed_wrap)

        # divide sum by amout of window_len, and resize to same format as input 'x'
        ret = resize(self.sum >> self.window_pow, size_res=x)
        return ret

    def model_main(self, inputs):
        taps = [1 / self.window_len] * self.window_len
        ret = np.convolve(inputs, taps, mode='full')
        return ret[:-self.window_len + 1]



import inspect
import textwrap
from pathlib import Path
from typing import List

from redbaron import RedBaron

from pyha.common.util import get_iterable
from pyha.conversion.converter import convert, file_header
from pyha.conversion.coupling import get_instance_vhdl_name, VHDLType
from pyha.conversion.extract_datamodel import DataModel
from pyha.conversion.top_generator import TopGenerator


def get_objects_rednode(obj):
    source_path = inspect.getsourcefile(type(obj))
    source = open(source_path).read()
    red_list = RedBaron(source)('class', name=obj.__class__.__name__)
    if len(red_list) != 1:
        raise MultipleNodesError('Found {} definitions of "{}" class'.
                                 format(len(red_list), obj.__class__.__name__))

    return red_list[0]


def get_conversion_datamodel(obj):
    red_node = get_objects_rednode(obj)
    datamodel = DataModel(obj)
    conv = convert(red_node, caller=None, datamodel=datamodel)
    return conv, datamodel


def get_conversion(obj):
    conv, _ = get_conversion_datamodel(obj)
    return conv


class MultipleNodesError(Exception):
    pass


class Conversion:
    """
    input: stimulated object
    outputs:
        *comonent vhdl files
        *top file
        *top input types
        *top output types
    """
    # collect all complex types in the design
    complex_types = []

    def __init__(self, obj, is_child=False):

        self.is_child = is_child
        self.obj = obj
        self.class_name = obj.__class__.__name__
        self.conv, self.datamodel = get_conversion_datamodel(obj)

        self.vhdl_conversion = str(self.conv)
        if not is_child:
            self.top_vhdl = TopGenerator(obj)
            Conversion.complex_types = []
        Conversion.complex_types.append(VHDLType.get_complex_vars())

        # recursively convert all child modules
        self.childs = []

        for x in self.obj._pyha_submodules:
            if isinstance(x, list):
                x = x[0]  # in case of submodules list -> only convert one
            self.childs.append(Conversion(x, is_child=True))

    @property
    def inputs(self) -> List[object]:
        # return [x.value if isinstance(x, Const) else x for x in self.top_vhdl.get_object_inputs()]
        return self.top_vhdl.get_object_inputs()

    @property
    def outputs(self) -> List[object]:
        return self.top_vhdl.get_object_return()

    def write_vhdl_files(self, base_dir: Path) -> List[Path]:

        # todo: makedirs
        paths = []
        for x in self.childs:
            paths.extend(x.write_vhdl_files(base_dir))  # recusion here

        # paths.append(base_dir / '{}.vhd'.format(self.class_name))
        fname = get_instance_vhdl_name(self.obj)
        paths.append(base_dir / f'{fname}.vhd')
        with paths[-1].open('w') as f:
            f.write(str(self.vhdl_conversion))

        # add top_generator file
        if not self.is_child:
            paths.append(base_dir / 'top.vhd')
            with paths[-1].open('w') as f:
                f.write(self.top_vhdl.make())

            paths.insert(0, base_dir / 'complex_types.vhd')
            with paths[0].open('w') as f:
                f.write(self.make_vhdl_complex_types())

        return paths


        # def get_objects_source_path(self, obj) -> str:
        #     return inspect.getsourcefile(type(obj))

    def make_vhdl_complex_types(self):
        template_none = textwrap.dedent("""\
            library ieee;
                use ieee.fixed_pkg.all;

            package ComplexTypes is
            end package;
            """)
        import numpy as np
        if len(np.array(self.complex_types).flatten()) == 0:
            return template_none

        template = textwrap.dedent("""\
            {FILE_HEADER}
            library ieee;
                use ieee.fixed_pkg.all;

            package ComplexTypes is
            {TYPES}
            end package;

            package body ComplexTypes is
            {FUNCTIONS}
            end package body;
            """)

        types = []
        functions = []
        for x in self.complex_types:
            for xx in get_iterable(x):
                new_type = xx.vhdl_type_define()
                new_function = xx.vhdl_init_function()
                if new_type not in types:
                    types.append(new_type)
                    functions.append(new_function)

        return template.format(FILE_HEADER=file_header(),
                               TYPES='\n'.join(x for x in types),
                               FUNCTIONS='\n'.join(x for x in functions))



import logging
import os
import shutil
import subprocess
import sys
from pathlib import Path

import numpy as np

import pyha
from pyha.common.sfix import Sfix, ComplexSfix

COCOTB_MAKEFILE_TEMPLATE = """
###############################################################################
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Potential Ventures Ltd,
#       SolarFlare Communications Inc nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL POTENTIAL VENTURES LTD BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
###############################################################################
include $(COCOTB)/makefiles/Makefile.inc
include $(COCOTB)/makefiles/Makefile.sim
"""


class CocotbAuto(object):
    def __init__(self, base_path, src, outputs, sim_folder='coco_sim'):
        self.logger = logging.getLogger(__name__)
        self.outputs = outputs
        self.src = src
        self.base_path = base_path
        self.sim_folder = sim_folder

        self.environment = os.environ.copy()
        self.default_assignments()

    def default_assignments(self):
        # self.environment['COCOTB'] = pyha.__path__[0] + '/../cocotb'

        # this line is called 'i hate cocotb'
        # ill throw my computer out of the window counter: 6
        self.environment["PYTHONHOME"] = str(Path(sys.executable).parent.parent)

        self.environment['SIM_BUILD'] = self.sim_folder
        self.environment['TOPLEVEL_LANG'] = 'vhdl'
        self.environment['SIM'] = 'ghdl'

        self.environment['GHDL_OPTIONS'] = '--std=08'  # TODO: push PR to cocotb

        if len(self.src) == 1:  # one file must be quartus netlist, need to simulate in 93 mode
            try:
                ghdl_path = Path(shutil.which('ghdl'))
            except:
                raise Exception('You dont have GHDL in PATH!')
            altera_libs = str(ghdl_path.parent.parent / 'lib/ghdl/altera')
            # altera_libs = pyha.__path__[0] + '/common/hdl/altera'
            self.environment[
                'GHDL_OPTIONS'] = '-P' + altera_libs + ' --ieee=synopsys --no-vital-checks'  # TODO: push PR to cocotb

        self.environment["PYTHONPATH"] = str(self.base_path)

        self.environment['TOPLEVEL'] = 'top'
        self.environment['MODULE'] = 'cocotb_simulation_top'

        self.environment['VHDL_SOURCES'] = ' '.join(str(x) for x in self.src)

        # copy cocotb simulation top file
        coco_py = pyha.__path__[0] + '/simulation/cocotb_simulation_top.py'
        shutil.copyfile(coco_py, str(self.base_path / Path(coco_py).name))
        self.environment['OUTPUT_VARIABLES'] = str(len(self.outputs))

    def run(self, *input_data):
        self.logger.info('Running COCOTB simulation....')
        # # convert all Sfix elements to 'integer' form
        input_data = np.vectorize(
            lambda x: x.fixed_value() if isinstance(x, (Sfix, ComplexSfix)) else x)(input_data)

        np.save(str(self.base_path / 'input.npy'), input_data)

        # write makefile template
        with (self.base_path / 'Makefile').open('w') as f:
            f.write(COCOTB_MAKEFILE_TEMPLATE)

        try:
            subprocess.run("make", env=self.environment, cwd=str(self.base_path), check=True)
            pass
        except subprocess.CalledProcessError as err:
            print('GHDL failed!')
            import os
            os._exit(-1)

        outp = np.load(str(self.base_path / 'output.npy'))
        outp = outp.astype(object)

        # convert 'integer' form back to Sfix
        outp = np.transpose(outp)

        def getSignedNumber(number, bitLength):
            # http://stackoverflow.com/questions/1375897/how-to-get-the-signed-integer-value-of-a-long-in-python
            mask = (2 ** bitLength) - 1
            if number & (1 << (bitLength - 1)):
                return number | ~mask
            else:
                return number & mask

        for i, row in enumerate(outp):
            for j, val in enumerate(row):
                if isinstance(self.outputs[i], bool):
                    outp[i][j] = bool(int(val))
                elif isinstance(self.outputs[i], int):
                    val = getSignedNumber(int(val, 2), 32)
                    outp[i][j] = val
                elif not isinstance(self.outputs[i], list):
                    val = getSignedNumber(int(val, 2), len(self.outputs[i]))

                if isinstance(self.outputs[i], Sfix):
                    outp[i][j] = (val * 2 ** self.outputs[i].right)
                elif isinstance(self.outputs[i], ComplexSfix):
                    val = int(val.real)
                    mask = (2 ** (self.outputs[i].bitwidth() // 2)) - 1
                    real = (val >> (self.outputs[i].bitwidth() // 2))
                    real = getSignedNumber(real & mask, self.outputs[i].bitwidth() // 2)
                    real *= 2 ** self.outputs[i].right

                    imag = getSignedNumber(val & mask, self.outputs[i].bitwidth() // 2)
                    imag *= 2 ** self.outputs[i].right
                    outp[i][j] = real + imag * 1j
                elif isinstance(self.outputs[i], list) and isinstance(self.outputs[i][0], bool):
                    v = np.array([bool(int(x)) for x in val])
                    outp[i][j] = v
                    pass

        outp = np.squeeze(outp)  # example [[1], [2], [3]] -> [1, 2, 3]
        outp = np.transpose(outp)

        return outp



from pyha.common.hwsim import HW
from pyha.components.blade_demod.bits_decode import DemodToPacket
from pyha.components.blade_demod.blade_adaptor import BladeToComplex
from pyha.components.blade_demod.blade_demod import DemodQuadMavg


class Phantom2ReceiverBlade(HW):
    """ Adds BladeToComplex before Phantom2Receiver """
    def __init__(self):
        self.blade_to_complex = BladeToComplex()
        self.recv = Phantom2Receiver()

        self._delay = self.recv._delay + self.blade_to_complex._delay

    def main(self, i, q):
        c = self.blade_to_complex.main(i, q)
        packet_part, valid = self.recv.main(c)
        return packet_part, valid

    def model_main(self, i, q):
        c = self.blade_to_complex.model_main(i, q)
        a = self.recv.model_main(c)
        return a


class Phantom2Receiver(HW):
    def __init__(self):
        self.demod = DemodQuadMavg(0.5, 16)
        self.packet = DemodToPacket()

        self._delay = self.demod._delay + self.packet._delay

    def main(self, c):
        demod = self.demod.main(c)
        packet_part, valid = self.packet.main(demod)
        return packet_part, valid

    def model_main(self, c):
        demod = self.demod.model_main(c)
        packet = self.packet.model_main(demod)
        return packet



import datetime
import logging
import textwrap
from contextlib import suppress
from enum import Enum

from parse import parse
from redbaron import NameNode, Node, EndlNode, DefNode, AssignmentNode, TupleNode, CommentNode, AssertNode
from redbaron.base_nodes import DotProxyList
from redbaron.nodes import AtomtrailersNode

import pyha
from pyha.common.hwsim import SKIP_FUNCTIONS, HW
from pyha.common.sfix import ComplexSfix
from pyha.common.sfix import Sfix
from pyha.common.util import get_iterable, tabber, escape_for_vhdl
from pyha.conversion.coupling import VHDLType, VHDLVariable, pytype_to_vhdl, list_reset
from pyha.conversion.coupling import get_instance_vhdl_name

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class ExceptionReturnFunctionCall(Exception):
    def __init__(self, red_node: Node):
        message = f'Trying to return something that is not an variable!\nLine: {red_node}'
        super().__init__(message)


# class ExceptionUnknownReturnVariable(Exception):
#     def __init__(self, datamodel: DataModel, red_node: Node):
#         message = textwrap.dedent("""\
#         Did not find returned variable in datamodel:
#         Line: {}
#         Datamodel:
#             {}
#         """).format(red_node, datamodel)
#         super().__init__(message)de


def file_header():
    template = '-- generated by pyha {} at {}'
    return template.format(pyha.__version__, datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'))


class NodeConv:
    def __init__(self, red_node, parent=None):
        self.red_node = red_node
        self.parent = parent
        self.target = None
        self.value = None
        self.first = None
        self.second = None
        self.test = None
        self.arguments = None
        self.name = None
        self.iterator = None

        for x in red_node._dict_keys:
            self.__dict__[x] = red_to_conv_hub(red_node.__dict__[x], caller=self)

        for x in red_node._list_keys:
            if 'format' not in x:
                self.__dict__[x] = []
                for xj in red_node.__dict__[x]:
                    if isinstance(xj, DefNode) and xj.name in SKIP_FUNCTIONS:
                        continue
                    self.__dict__[x].append(red_to_conv_hub(xj, caller=self))

        # FIXME: possible bug, need to process strings?
        for x in red_node._str_keys:
            self.__dict__[x] = red_node.__dict__[x]

    # def __iter__(self):
    #     return iter(self.nodes)

    def __str__(self):
        return str(self.red_node)


class NameNodeConv(NodeConv):
    def __str__(self):
        return escape_for_vhdl(self.red_node.value)


class AtomtrailersNodeConv(NodeConv):
    def is_function_call(self):
        return any(isinstance(x, CallNodeConv) for x in self.value)

    def __str__(self):
        ret = ''
        for i, x in enumerate(self.value):
            # add '.' infront if NameNode
            new = '.{}' if isinstance(x, NameNodeConv) and i != 0 else '{}'
            ret += new.format(x)

        return ret


class TupleNodeConv(NodeConv):
    def __iter__(self):
        return iter(self.value)

    def __str__(self):
        return ','.join(str(x) for x in self.value)


class AssignmentNodeConv(NodeConv):
    def __str__(self):
        r = f'{self.target} := {self.value};'
        if isinstance(self.red_node.target, TupleNode) or isinstance(self.red_node.value, TupleNode):
            raise Exception(f'{r} -> multi assignment not supported!')

        if self.red_node.operator != '':
            raise Exception('{} -> cannot convert +=, -=, /=, *= :(')
        return r


class ReturnNodeConv(NodeConv):
    def __str__(self):
        for x in get_iterable(self.value):
            if isinstance(x, AtomtrailersNodeConv) and x.is_function_call():
                raise ExceptionReturnFunctionCall(self.red_node)
            elif not isinstance(x, (NameNodeConv, AtomtrailersNodeConv, IntNodeConv)):
                raise ExceptionReturnFunctionCall(self.red_node)

        str_ret = [f'ret_{i} := {ret};' for i, ret in enumerate(get_iterable(self.value))]
        str_ret += ['return;']
        return '\n'.join(str_ret)


class ComparisonNodeConv(NodeConv):
    def __str__(self):
        return f'{self.first} {self.value} {self.second}'


class BinaryOperatorNodeConv(ComparisonNodeConv):
    def __str__(self):

        # test if we are dealing with array appending ([a] + b)
        if self.value == '+':
            if isinstance(self.first, ListNodeConv) or isinstance(self.second, ListNodeConv):
                self.value = '&'
        elif self.value == '>>':
            self.value = 'sra'
        elif self.value == '<<':
            self.value = 'sla'
        elif self.value == '&':
            self.value = 'and'
        elif self.value == '|':
            self.value = 'or'
        elif self.value == '^':
            self.value = 'xor'

        return f'{self.first} {self.value} {self.second}'


class BooleanOperatorNodeConv(ComparisonNodeConv):
    pass


class AssociativeParenthesisNodeConv(NodeConv):
    def __str__(self):
        return f'({self.value})'


class ComparisonOperatorNodeConv(NodeConv):
    def __str__(self):
        if self.first == '==':
            return '='
        elif self.first == '!=':
            return '/='
        else:
            return super().__str__()


class IfelseblockNodeConv(NodeConv):
    def __str__(self):
        body = '\n'.join(str(x) for x in self.value)
        return body + '\nend if;'


class IfNodeConv(NodeConv):
    def __str__(self):
        body = '\n'.join(tabber(str(x)) for x in self.value)
        return f'if {self.test} then\n{body}'


class ElseNodeConv(NodeConv):
    def __str__(self):
        body = '\n'.join(tabber(str(x)) for x in self.value)
        return f'else\n{body}'


class ElifNodeConv(NodeConv):
    def __str__(self):
        body = '\n'.join(tabber(str(x)) for x in self.value)
        return f'elsif {self.test} then\n{body}'


class DefNodeConv(NodeConv):
    def __init__(self, red_node, parent=None):
        super().__init__(red_node, parent)
        self.name = escape_for_vhdl(self.name)

        # collect multiline comment
        self.multiline_comment = ''
        if isinstance(self.value[0], StringNodeConv):
            self.multiline_comment = str(self.value[0])
            del self.value[0]

        # remove last line,i f it is \n
        if isinstance(self.value[-1], EndlNodeConv):
            del self.value[-1]

        self.arguments.extend(self.infer_return_arguments())
        self.variables = self.infer_variables()

    def infer_return_arguments(self):
        try:
            rets = self.red_node('return')[0]
        except IndexError:
            return []

        def get_type(i: int, red):
            name = escape_for_vhdl('ret_' + str(i))
            return VHDLType(name, port_direction='out', red_node=red)

        # atomtrailers return len() > 1 for one return element
        if isinstance(rets.value, AtomtrailersNode) or len(rets.value) == 1:
            return [get_type(0, rets.value)]

        return [get_type(i, x) for i, x in enumerate(rets.value)]

    def infer_variables(self):
        assigns = self.red_node.value('assign')

        variables = []
        for x in assigns:
            if isinstance(x.target, NameNode):
                variables.append(VHDLVariable(NameNodeConv(red_node=x.target), red_node=x))
            elif isinstance(x.target, TupleNode):
                for node in x.target:
                    # variables.append(VHDLVariable(NameNodeConv(red_node=node), red_node=x))
                    variables.append(VHDLVariable(str(node), red_node=node))

        call_args = self.red_node.value('call_argument')
        call_args = [x for x in call_args if str(x)[:4] == 'ret_']
        for x in call_args:
            if isinstance(x.value, AtomtrailersNode) and str(x.value[0]) == 'self':
                continue
            getitem = x.value.getitem
            if getitem is not None:
                variables.append(VHDLVariable(escape_for_vhdl(str(getitem.previous)), red_node=x.value))
            else:
                variables.append(VHDLVariable(escape_for_vhdl(str(x.value)), red_node=x.value))

        remove_duplicates = {str(x.name): x for x in variables}
        variables = remove_duplicates.values()

        # remove variables that are actually arguments
        args = [str(x.target.name) for x in self.arguments]
        return [x for x in variables if str(x.name) not in args]

    def get_prototype(self):
        template = textwrap.dedent("""\
            {MULTILINE_COMMENT}
            procedure {NAME}{ARGUMENTS};""")
        sockets = {'NAME': self.name, 'MULTILINE_COMMENT': self.multiline_comment}

        sockets['ARGUMENTS'] = ''
        if len(self.arguments):
            sockets['ARGUMENTS'] = '(' + '; '.join(str(x) for x in self.arguments) + ')'

        return template.format(**sockets)

    def __str__(self):
        template = textwrap.dedent("""\
            {MULTILINE_COMMENT}
            procedure {NAME}{ARGUMENTS} is
            {VARIABLES}
            begin
            {BODY}
            end procedure;""")

        sockets = {'NAME': self.name, 'MULTILINE_COMMENT': self.multiline_comment}

        sockets['ARGUMENTS'] = ''
        if len(self.arguments):
            sockets['ARGUMENTS'] = '(' + '; '.join(str(x) for x in self.arguments) + ')'

        sockets['VARIABLES'] = ''
        if len(self.variables):
            sockets['VARIABLES'] = '\n'.join(tabber(str(x)) for x in self.variables)

        sockets['BODY'] = '\n'.join(tabber(str(x)) for x in self.value)
        return template.format(**sockets)


class DefArgumentNodeConv(NodeConv):
    def __init__(self, red_node, parent=None):
        super().__init__(red_node, parent)
        self.target = VHDLType(name=self.target, red_node=red_node, value=self.value)
        self.name = self.target.name

    def __str__(self):
        return str(self.target)


class PassNodeConv(NodeConv):
    def __str__(self):
        return ''


class CallNodeConv(NodeConv):
    def __str__(self):
        base = '(' + ', '.join(str(x) for x in self.value) + ')'
        if str(self.red_node.parent) == 'make_self(self_reg, self)':
            return base + ';'  # fixes some random bug

        is_assign = self.red_node.parent_find('assign')
        if not is_assign and isinstance(self.red_node.next_recursive, EndlNode):
            base += ';'
        return base


class CallArgumentNodeConv(NodeConv):
    def __str__(self):
        # transform keyword arguments
        # change = to =>
        if self.target is not None:
            return f'{self.target}=>{self.value}'

        return str(self.value)

    pass


class IntNodeConv(NodeConv):
    pass


class FloatNodeConv(NodeConv):
    pass


class UnitaryOperatorNodeConv(NodeConv):
    pass


class AssertNodeConv(NodeConv):
    def __str__(self):
        return '--' + super().__str__()


class PrintNodeConv(NodeConv):
    def __str__(self):
        if isinstance(self.red_node.value[0], TupleNode):
            raise Exception(f'{self.red_node} -> print only supported with one Sfix argument!')
        return f"report to_string({self.red_node.value[0].value});"
        return f"report to_string(to_real({self.red_node.value[0].value}));"


class ListNodeConv(NodeConv):
    def __str__(self):
        if len(self.value) == 1:
            return str(self.value[0])  # [a] -> a
        else:
            ret = f'({", ".join(str(x) for x in self.value)})'
            return ret


class EndlNodeConv(NodeConv):
    def __str__(self):
        if isinstance(self.red_node.previous_rendered, CommentNode):
            return '--' + str(self.red_node.previous_rendered)[1:]
        return ''


class HexaNodeConv(NodeConv):
    def __str__(self):
        return f'16#{self.value[2:]}#'


class CommentNodeConv(NodeConv):
    def __str__(self):
        return '--' + self.value[1:]


class StringNodeConv(NodeConv):
    """ Multiline comments come here """

    def __str__(self):
        if self.value[:3] == '"""' and self.value[-3:] == '"""':
            r = [x.strip() for x in self.value[3:-3].splitlines()]
            r = '\n-- '.join(x for x in r if x != '')
            return '-- ' + r

        return self.value[1:]


# this is mostly array indexing
class GetitemNodeConv(NodeConv):
    # turn python [] indexing to () indexing

    def get_index_target(self):
        ret = ''
        for x in self.parent.value:
            if x is self:
                break
            ret += '.' + str(x)
        return ret[1:]

    def is_negative_indexing(self, obj):
        return isinstance(obj, UnitaryOperatorNodeConv) and int(str(obj)) < 0

    def __str__(self):
        if self.is_negative_indexing(self.value):
            target = self.get_index_target()
            return f"({target}'length{self.value})"

        return f'({self.value})'


class SliceNodeConv(GetitemNodeConv):
    def get_index_target(self):
        return '.'.join(str(x) for x in self.parent.parent.value[:-1])

    # Example: [0:5] -> (0 to 4)
    # x[0:-1] -> x(0 to x'high-1)
    def __str__(self):
        if self.upper is None:
            upper = f"{self.get_index_target()}'high"
        else:
            # vhdl includes upper limit, subtract one to get same behaviour as in python
            # upper = int(str(self.upper)) - 1
            upper = f'({self.upper})-1'

        if self.is_negative_indexing(self.upper):
            target = self.get_index_target()
            upper = f"{target}'high{self.upper}"

        lower = 0 if self.lower is None else self.lower
        return f'{lower} to {upper}'


class ForNodeConv(NodeConv):
    def __str__(self):
        template = textwrap.dedent("""\
                for {ITERATOR} in {RANGE} loop
                {BODY}
                end loop;""")

        sockets = {'ITERATOR': str(self.iterator)}
        sockets['RANGE'] = self.range_to_vhdl(str(self.target))
        sockets['BODY'] = '\n'.join(tabber(str(x)) for x in self.value)
        return template.format(**sockets)

    def range_to_vhdl(self, pyrange):
        # this for was transforemed by 'redbaron_pyfor_to_vhdl'
        if str(self.iterator) == '\\_i_\\':
            return f"{pyrange}'range"

        range_len_pattern = parse('\\range\\(len({}))', pyrange)
        if range_len_pattern is not None:
            return range_len_pattern[0] + "'range"
        else:
            range_pattern = parse('\\range\\({})', pyrange)
            if range_pattern is not None:
                two_args = parse('{},{}', range_pattern[0])
                if two_args is not None:
                    return f'{two_args[0].strip()} to ({two_args[1].strip()}) - 1'
                else:
                    len = parse('len({}){}', range_pattern[0])
                    if len is not None:
                        return f"0 to ({len[0]}'length{len[1]}) - 1"
                    return f'0 to ({range_pattern[0]}) - 1'

        # at this point range was not:
        # range(len(x))
        # range(x)
        # range(x, y)
        # assume
        assert 0


class ClassNodeConv(NodeConv):
    """ This relies heavily on datamodel """

    def __init__(self, red_node, parent=None):
        super().__init__(red_node, parent)

        # collect multiline comment
        self.multiline_comment = ''
        if len(self.value) and isinstance(self.value[0], StringNodeConv):
            self.multiline_comment = str(self.value[0])
            del self.value[0]

    def get_imports(self):
        return textwrap.dedent("""\
            library ieee;
                use ieee.std_logic_1164.all;
                use ieee.numeric_std.all;
                use ieee.fixed_float_types.all;
                use ieee.fixed_pkg.all;
                use ieee.math_real.all;

            library work;
                use work.ComplexTypes.all;
                use work.PyhaUtil.all;
                use work.all;""")

    def get_reset_self_prototype(self):
        return 'procedure \\_pyha_reset_self\\(self: inout self_t);'

    def get_reset_self(self):
        template = textwrap.dedent("""\
        procedure \\_pyha_reset_self\\(self: inout self_t) is
        begin
        {DATA}
            \\_pyha_update_self\\(self);
        end procedure;""")

        variables = VHDLType.get_reset()

        sockets = {'DATA': ''}
        sockets['DATA'] += ('\n'.join(tabber(x) for x in variables))
        return template.format(**sockets)

    def get_update_self_prototype(self):
        return 'procedure \\_pyha_update_self\\(self: inout self_t);'

    def get_update_self(self):
        template = textwrap.dedent("""\
        procedure \\_pyha_update_self\\(self: inout self_t) is
        begin
        {DATA}
            \\_pyha_constants_self\\(self);
        end procedure;""")

        sockets = {'DATA': ''}
        lines = []
        for x in VHDLType.get_self():
            var_name = x.name
            var_value = x.variable
            if isinstance(var_value, HW):
                lines.append(f'{get_instance_vhdl_name(var_value)}.\\_pyha_update_self\\(self.{var_name});')
            elif isinstance(var_value, list) and isinstance(var_value[0], HW):
                for i in range(len(var_value)):
                    lines.append(f'{get_instance_vhdl_name(var_value[0])}.\\_pyha_update_self\\(self.{var_name}({i}));')
            else:
                lines.append(f'self.{var_name} := self.\\next\\.{var_name};')

        sockets['DATA'] += ('\n'.join(tabber(x) for x in lines))

        return template.format(**sockets)

    def get_constants_self_prototype(self):
        return 'procedure \\_pyha_constants_self\\(self: inout self_t);'

    def get_constants_self(self):
        template = textwrap.dedent("""\
        procedure \\_pyha_constants_self\\(self: inout self_t) is
        begin
        {CONSTANTS}
        {SUBMODULES}
        end procedure;""")

        sockets = {'CONSTANTS': '', 'SUBMODULES': ''}

        # call constants self for each submodules, quartus wants this
        lines = []
        for x in VHDLType.get_self():
            var_name = x.name
            var_value = x.variable
            if isinstance(var_value, HW):
                lines.append(f'{get_instance_vhdl_name(var_value)}.\\_pyha_constants_self\\(self.{var_name});')
            elif isinstance(var_value, list) and isinstance(var_value[0], HW):
                for i in range(len(var_value)):
                    lines.append(
                        f'{get_instance_vhdl_name(var_value[0])}.\\_pyha_constants_self\\(self.{var_name}({i}));')

        sockets['SUBMODULES'] += ('\n'.join(tabber(x) for x in lines))

        const = VHDLType.get_constants()
        if len(const):
            const_str = []
            for var in const:
                value = var.variable
                if isinstance(value, Enum):
                    const_str += [f'self.{var.name} := {value.name};']
                elif isinstance(value, (Sfix, ComplexSfix)):
                    const_str += [f'self.{var.name} := {value.vhdl_reset()};']
                elif isinstance(value, list):
                    const_str += ['self.' + list_reset('', var.name, value)]
                else:
                    const_str += [f'self.{var.name} := {value};']

            sockets['CONSTANTS'] += ('\n'.join(tabber(x) for x in const_str))

        return template.format(**sockets)

    def get_datamodel(self):
        template = textwrap.dedent("""\
            type next_t is record
            {DATA}
            end record;

            type self_t is record
            {CONSTANTS}
            {DATA}
                \\next\\: next_t;
            end record;""")
        sockets = {'DATA': '', 'CONSTANTS': ''}
        sockets['DATA'] += ('\n'.join(tabber(str(x) + ';') for x in VHDLType.get_self()))
        const = VHDLType.get_constants()
        if len(const):
            sockets['CONSTANTS'] = '    -- constants\n'
            sockets['CONSTANTS'] += ('\n'.join(tabber(str(x) + ';') for x in const))
            sockets['CONSTANTS'] += '\n'
        return template.format(**sockets)

    def get_typedefs(self):
        typedefs = []
        for val in VHDLType.get_typedef_vars():
            assert type(val) is list
            name = pytype_to_vhdl(val)
            name = name[:name.find('(')]  # cut array size

            type_name = pytype_to_vhdl(val[0])
            if isinstance(val[0], HW):
                type_name += '.self_t'
            new_tp = f'type {name} is array (natural range <>) of {type_name};'
            if new_tp not in typedefs:
                typedefs.append(new_tp)

        return typedefs

    def get_enumdefs(self):
        vals = []
        for val in VHDLType.get_enum_vars():
            name = pytype_to_vhdl(val)

            types = ','.join([x.name for x in type(val)])
            new_tp = f'type {name} is ({types});'
            if new_tp not in vals:
                vals.append(new_tp)

        return vals

    def get_name(self):
        return VHDLType.get_self_vhdl_name()

    def get_headers(self):
        ret = self.get_constants_self_prototype() + '\n\n'
        ret += self.get_reset_self_prototype() + '\n\n'
        ret += self.get_update_self_prototype() + '\n\n'
        ret += '\n\n'.join(x.get_prototype() for x in self.value if isinstance(x, DefNodeConv))
        return ret

    def get_function(self, name):
        f = [x for x in self.value if str(x.name) == name]
        return str(f[0])

    def get_package_header(self):
        template = textwrap.dedent("""\
            {MULTILINE_COMMENT}
            package {NAME} is
            {ENUMDEFS}
            {TYPEDEFS}

            {SELF_T}

            {FUNC_HEADERS}
            end package;""")

        sockets = {}
        sockets['MULTILINE_COMMENT'] = self.multiline_comment
        sockets['NAME'] = self.get_name()
        sockets['ENUMDEFS'] = '\n'.join(tabber(x) for x in self.get_enumdefs())
        sockets['TYPEDEFS'] = '\n'.join(tabber(x) for x in self.get_typedefs())
        sockets['SELF_T'] = tabber(self.get_datamodel())

        sockets['FUNC_HEADERS'] = tabber(self.get_headers())

        return template.format(**sockets)

    def get_package_body(self):
        template = textwrap.dedent("""\
            package body {NAME} is
            {CONSTANT_SELF}

            {RESET_SELF}

            {UPDATE_SELF}

            {OTHER_FUNCTIONS}
            end package body;""")

        sockets = {}
        sockets['NAME'] = self.get_name()

        sockets['CONSTANT_SELF'] = tabber(self.get_constants_self())
        sockets['RESET_SELF'] = tabber(self.get_reset_self())
        sockets['UPDATE_SELF'] = tabber(self.get_update_self())
        sockets['OTHER_FUNCTIONS'] = '\n\n'.join(tabber(str(x)) for x in self.value)

        return template.format(**sockets)

    def __str__(self):
        template = textwrap.dedent("""\
            {FILE_HEADER}
            {IMPORTS}

            {PACKAGE_HEADER}

            {PACKAGE_BODY}
            """)

        sockets = {}
        sockets['FILE_HEADER'] = file_header()
        sockets['IMPORTS'] = self.get_imports()
        sockets['PACKAGE_HEADER'] = self.get_package_header()
        sockets['PACKAGE_BODY'] = self.get_package_body()
        return template.format(**sockets)


def red_to_conv_hub(red: Node, caller):
    """ Convert RedBaron class to conversion class
    For example: red:NameNode returns NameNodeConv class
    """
    import sys

    red_type = red.__class__.__name__
    try:
        cls = getattr(sys.modules[__name__], red_type + 'Conv')
    except AttributeError:
        if red_type == 'NoneType':
            return None
        raise

    return cls(red_node=red, parent=caller)


def convert(red: Node, caller=None, datamodel=None):
    from pyha.conversion.extract_datamodel import DataModel
    assert type(caller) is not DataModel
    VHDLType.set_datamodel(datamodel)

    # delete __init__, not converting this
    with suppress(AttributeError):
        f = red.find('def', name='__init__')
        f.parent.remove(f)

    # delete model_main, not converting this
    with suppress(AttributeError):
        f = red.find('def', name='model_main')
        f.parent.remove(f)
    if datamodel is not None:
        red = redbaron_enum_to_vhdl(red)
    red = redbaron_pyfor_to_vhdl(red)
    red = redbaron_pycall_returns_to_vhdl(red)
    red = redbaron_pycall_to_vhdl(red)

    conv = red_to_conv_hub(red, caller)  # converts all nodes

    return conv


#################### FUNCTIONS THAT MODIFY REDBARON AST #############
#####################################################################
#####################################################################
#####################################################################
#####################################################################
#####################################################################

def redbaron_enum_to_vhdl(red_node):
    """ In python Enums must be referenced by type: EnumType.ENUMVALUE
    VHDL does not allow  this, only ENUMVALUE must be written"""
    enums = VHDLType.get_enum_vars()
    for x in enums:
        type_name = type(x).__name__
        red_names = red_node.find_all('atomtrailers', value=lambda x: x[0].value == type_name)
        for i, node in enumerate(red_names):
            red_names[i].replace(node[1])

    return red_node


def redbaron_pycall_to_vhdl(red_node):
    """
    Main work is to add 'self' argument to function call
    self.d(a) -> d(self, a)

    If function owner is not exactly 'self' then 'unknown_type' is prepended.
    self.next.moving_average.main(x) -> unknown_type.main(self.next.moving_average, x)

    self.d(a) -> d(self, a)
    self.next.d(a) -> d(self.next, a)
    local.d() -> type.d(local)
    self.local.d() -> type.d(self.local)

    """

    def modify_call(red_node):
        call_args = red_node.find('call')
        i = call_args.previous.index_on_parent
        if i == 0:
            return red_node  # input is something like a()

        if isinstance(red_node.parent, AssertNode):
            return red_node
        prefix = red_node.copy()
        del prefix[i:]
        del red_node[:i]

        # this happens when 'redbaron_pyfor_to_vhdl' does some node replacements
        if isinstance(prefix.value, DotProxyList) and len(prefix) == 1:
            prefix = prefix[0]

        call_args.insert(0, prefix)
        if prefix.dumps() not in ['self', 'self.next']:
            v = VHDLType(str(prefix[-1]), red_node=prefix)
            red_node.insert(0, v.var_type)

    atoms = red_node.find_all('atomtrailers')
    for i, x in enumerate(atoms):
        if x.call is not None:
            modify_call(x)

    return red_node


def redbaron_pycall_returns_to_vhdl(red_node):
    """
    Convert function calls, that return into variable into VHDL format.
    b = self.a(a) ->
        self.a(a, ret_0=b)

    self.next.b[0], self.next.b[1] = self.a(self.a) ->
        self.a(self.a, ret_0=self.next.b[0], ret_1=self.next.b[1])

    """

    def modify_call(x: AssignmentNode):
        try:
            if str(x.value[0]) != 'self':  # most likely call to 'resize' no operatons needed
                if str(x.value[0][0]) != 'self':  # this is some shit that happnes after 'for' transforms
                    return x
        except:
            return x

        call = x.call
        if len(x.target) == 1 or isinstance(x.target, AtomtrailersNode):
            call.append(str(x.target))
            call.value[-1].target = 'ret_0'
        else:
            for j, argx in enumerate(x.target):
                call.append(str(argx))
                call.value[-1].target = f'ret_{j}'
        return x.value

    assigns = red_node.find_all('assign')
    for x in assigns:
        if x.call is not None:
            new = modify_call(x.copy())
            x.replace(new)
    return red_node


def redbaron_pyfor_to_vhdl(red_node):
    def modify_for(red_node):
        # if for range contains call to 'range' -> skip
        with suppress(Exception):
            if red_node.target('call')[0].previous.value == 'range':
                return red_node

        frange = red_node.target
        ite = red_node.iterator

        red_node(ite.__class__.__name__, value=ite.value) \
            .map(lambda x: x.replace(f'{frange}[_i_]'))

        red_node.iterator = '_i_'
        return red_node

    fors = red_node.find_all('for')
    for x in fors:
        modify_for(x)

    return red_node



import logging
import textwrap

import numpy as np

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

fixed_truncate = 'fixed_truncate'
fixed_round = 'fixed_round'

fixed_saturate = 'fixed_saturate'
fixed_wrap = 'fixed_wrap'
fixed_wrap_impossible = 'fixed_wrap_impossible'


class ComplexSfix:
    """
    Use real and imag members to access underlying Sfix elements.

    :param val:
    :param left: left bound for both components
    :param right: right bound for both components
    :param overflow_style: fixed_saturate(default) or fixed_wrap

    >>> a = ComplexSfix(0.45 + 0.88j, left=0, right=-17)
    >>> a
    0.45+0.88j [0:-17]
    >>> a.real
    0.4499969482421875 [0:-17]
    >>> a.imag
    0.8799972534179688 [0:-17]

    Another way to construct it:

    >>> a = Sfix(-0.5, 0, -17)
    >>> b = Sfix(0.5, 0, -17)
    >>> ComplexSfix(a, b)
    -0.50+0.50j [0:-17]


    """
    def __init__(self, val=0.0 + 0.0j, left=0, right=0, overflow_style=fixed_saturate):
        if type(val) is Sfix and type(left) is Sfix:
            self.init_val = val.init_val + left.init_val * 1j
            self.real = val
            self.imag = left
        else:
            self.init_val = val
            self.real = Sfix(val.real, left, right, overflow_style=overflow_style)
            self.imag = Sfix(val.imag, left, right, overflow_style=overflow_style)

    @property
    def left(self):
        assert self.real.left == self.imag.left
        return self.real.left

    @property
    def right(self):
        assert self.real.right == self.imag.right
        return self.real.right

    @property
    def val(self):
        return self.real.val + self.imag.val * 1j

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    def __call__(self, x):
        return ComplexSfix(x, self.left, self.right)

    def __str__(self):
        return f'{self.real.val:.2f}{"" if self.imag.val < 0.0 else "+"}{self.imag.val:.2f}j [{self.left}:{self.right}]'

    def __repr__(self):
        return str(self)

    def has_same_bounds(self, other):
        if self.left == other.left and self.right == other.right:
            return True
        return False

    def bitwidth(self):
        return (self.left + abs(self.right)) * 2 + 2

    def __len__(self):
        return self.bitwidth()

    def to_stdlogic(self):
        return f'std_logic_vector({self.bitwidth() - 1} downto 0)'

    def vhdl_reset(self):
        return f'(real=>{self.real.vhdl_reset()}, imag=>{self.imag.vhdl_reset()})'

    def fixed_value(self):
        assert self.bitwidth() <= 64  # must fit into numpy int, this is cocotb related?
        real = self.real.fixed_value()
        imag = self.imag.fixed_value()
        mask = (2 ** (self.bitwidth() // 2)) - 1
        return ((real & mask) << (self.bitwidth() // 2)) | (imag & mask)

    def vhdl_type_name(self):
        from pyha.conversion.coupling import pytype_to_vhdl
        return pytype_to_vhdl(self)

    def vhdl_type_define(self):
        dtype = f'sfixed({self.left} downto {self.right})'
        return textwrap.dedent(f"""\
            type {self.vhdl_type_name()} is record
                real: {dtype};
                imag: {dtype};
            end record;
            function ComplexSfix(a, b: sfixed({self.left} downto {self.right})) return {self.vhdl_type_name()};
            """)

    def vhdl_init_function(self):
        return textwrap.dedent(f"""\
            function ComplexSfix(a, b: sfixed({self.left} downto {self.right})) return {self.vhdl_type_name()} is
            begin
                return (a, b);
            end function;
            """)


# TODO: Verify stuff against VHDL library
class Sfix:
    """
    Signed fixed point type, like to_sfixed() in VHDL. Basic arithmetic operations
    are defined for this class.

    More info: https://www.dsprelated.com/showarticle/139.php

    :param val: initial value
    :param left: bits for integer part.
    :param right: bits for fractional part. This is negative number.
    :param init_only: internal use only
    :param overflow_style: fixed_saturate(default) or fixed_wrap
    :param round_style: fixed_round(default) or fixed_truncate

    >>> Sfix(0.123, left=0, right=-17)
    0.1230010986328125 [0:-17]
    >>> Sfix(0.123, left=0, right=-7)
    0.125 [0:-7]

    >>> Sfix(2.5, left=0, right=-17)
    WARNING:pyha.common.sfix:Saturation 2.5 -> 0.9999923706054688
    0.9999923706054688 [0:-17]
    >>> Sfix(2.5, left=1, right=-17)
    WARNING:pyha.common.sfix:Saturation 2.5 -> 1.9999923706054688
    1.9999923706054688 [1:-17]
    >>> Sfix(2.5, left=2, right=-17)
    2.5 [2:-17]

    """
    # original idea was to use float for internal computations, now it has turned out that\
    # it is hard to match VHDL fixed point library outputs, thus in the future it may be better
    # to implement stuff as integer arithmetic

    # Disables all quantization and saturating stuff
    _float_mode = False

    # TODO: finish this, there is unit test for it
    # @staticmethod
    # def auto_size(val, bits):
    #     """
    #     Find optimal sfixed format for value
    #
    #     :param val: May be list, then format is optimal including all elements
    #     :param bits:
    #     :return:
    #     """
    #     # FIXME: int_bits possibly not correct, since we cannot reporesent max positive value, actuall maximum value is (max_int) - (frac_min)
    #     # calculates for signed type
    #     if type(val) is list:
    #         maxabs = max(abs(x) for x in val)
    #         int_bits = np.floor(np.log2(np.abs(maxabs))) + 1
    #         fract_bits = -bits + int_bits + 1
    #         ret = [Sfix(x, int_bits, fract_bits) for x in val]
    #         return [Sfix(x, int_bits, fract_bits) for x in val]
    #     else:
    #         int_bits = np.floor(np.log2(np.abs(val))) + 1
    #         fract_bits = -bits + int_bits + 1
    #         return Sfix(val, int_bits, fract_bits)

    @staticmethod
    def set_float_mode(x):
        """
        Can be used to turn off all quantization effects, useful for debugging.

        :param x: True/False
        """
        Sfix._float_mode = x

    def __init__(self, val=0.0, left=0, right=0, init_only=False, overflow_style=fixed_saturate,
                 round_style=fixed_round):

        self.round_style = round_style
        self.overflow_style = overflow_style

        val = float(val)
        if type(val) not in [float, int]:
            raise Exception('Value must be float or int!')

        if isinstance(left, Sfix):
            self.right = left.right
            self.left = left.left
        else:
            self.right = right
            self.left = left

        assert self.left >= self.right
        self.val = val
        self.init_val = val

        # FIXME: This sucks, init should not call these anyways, make to_sfixed function
        if init_only or Sfix._float_mode:
            return

        if overflow_style is fixed_saturate:
            if self.overflows() and overflow_style:
                self.saturate()
            else:
                self.quantize()
        elif overflow_style in (fixed_wrap, fixed_wrap_impossible):  # TODO: add tests
            self.quantize()
            self.wrap()
        else:
            raise Exception('Wtf')

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    # FIXME: THESE ARE FUCKED UP
    def max_representable(self):
        if self.left < 0:
            # FIXME: I am not sure how to handle this when negative index
            # assert 0

            if self.right == 0:
                return 2 ** self.left
            else:
                return 2 ** self.left - 2 ** self.right
        else:
            return 2 ** self.left - 2 ** self.right

    def min_representable(self):
        return -2 ** self.left

    def overflows(self):
        return self.val < self.min_representable() or \
               self.val > self.max_representable()

    def is_lazy_init(self):
        return self.left == 0 and self.right == 0

    def saturate(self):
        old = self.val
        if self.val > self.max_representable():
            self.val = self.max_representable()
        elif self.val < self.min_representable():
            self.val = self.min_representable()
        else:
            assert False
        if not self.is_lazy_init() and not old == 1.0:
            logger.warning(f'Saturation {old} -> {self.val}')

    def wrap(self):
        if self.overflow_style is fixed_wrap_impossible:
            Exception('Wrap happened for "fixed_wrap_impossible"')

        fmin = self.min_representable()
        fmax = 2 ** self.left  # no need to substract minimal step, 0.9998... -> 1.0 will still be wrapped as max bit pattern
        self.val = (self.val - fmin) % (fmax - fmin) + fmin

    def quantize(self):
        fix = self.val / 2 ** self.right

        if self.round_style is fixed_round:
            fix = round(fix)
            # fix = np.round(fix)
        else:
            fix = int(fix)

        self.val = fix * 2 ** self.right

    # TODO: test, rounding not needed?
    def fixed_value(self):
        return int(np.round(self.val / 2 ** self.right))

    # def from_fixed(self, val, right):
    #     return (val * 2 ** self.outputs[i].right)

    def __str__(self):
        return f'{str(self.val)} [{self.left}:{self.right}]'

    def __repr__(self):
        return self.__str__()

    def __float__(self):
        return float(self.val)

    def resize(self, left=0, right=0, type=None, overflow_style=fixed_saturate, round_style=fixed_round):
        if type is not None:  # TODO: add tests
            left = type.left
            right = type.right
        return Sfix(self.val, left, right, overflow_style=overflow_style, round_style=round_style)

    def __add__(self, other):
        if type(other) == float:
            other = Sfix(other, self.left, self.right)
        return Sfix(self.val + other.val,
                    max(self.left, other.left) + 1,
                    min(self.right, other.right),
                    init_only=True)

    def __sub__(self, other):
        if type(other) == float:
            other = Sfix(other, self.left, self.right)
        return Sfix(self.val - other.val,
                    max(self.left, other.left) + 1,
                    min(self.right, other.right),
                    init_only=True)

    def __mul__(self, other):
        if type(other) == float:
            other = Sfix(other, self.left, self.right)
        return Sfix(self.val * other.val,
                    self.left + other.left + 1,
                    self.right + other.right,
                    init_only=True)

    def sign_bit(self):
        s = np.sign(self.val)
        if s in [0, 1]:
            return 0
        return 1

    def __rshift__(self, other):
        # todo: in float mode this should not lose precison
        o = int(self.val / 2 ** self.right)
        o = (o >> other) * 2 ** self.right
        return Sfix(o,
                    self.left,
                    self.right,
                    init_only=True)

    def __lshift__(self, other):
        # todo: in float mode this should not lose precison
        o = int(self.val / 2 ** self.right)
        o = (o << other) * 2 ** self.right
        return Sfix(o,
                    self.left,
                    self.right,
                    overflow_style=fixed_wrap)

    def __abs__(self):
        return Sfix(abs(self.val),
                    self.left + 1,
                    self.right,
                    init_only=True)

    # TODO: add tests
    def __lt__(self, other):
        return bool(self.val < other)

    # TODO: add tests
    def __gt__(self, other):
        return bool(self.val > other)

    # TODO: add tests
    def __neg__(self):
        return Sfix(-self.val,
                    self.left + 1,
                    self.right,
                    init_only=True)

    # TODO: add tests
    def __len__(self):
        assert self.left >= 0
        return -self.right + self.left + 1

    def __call__(self, x: float):
        return Sfix(x, self.left, self.right)

    def to_stdlogic(self):
        return f'std_logic_vector({self.left + abs(self.right)} downto 0)'

    def vhdl_reset(self):
        return f'Sfix({self.init_val}, {self.left}, {self.right})'


def resize(fix, left_index=0, right_index=0, size_res=None, overflow_style=fixed_saturate, round_style=fixed_round):
    """
    Resize fixed point number.

    :param fix: Sfix object to resize
    :param left_index: new left bound
    :param right_index: new right bound
    :param size_res: provide another Sfix object as size reference
    :param overflow_style: fixed_saturate(default) or fixed_wrap
    :param round_style: fixed_round(default) or fixed_truncate
    :return: New resized Sfix object

    >>> a = Sfix(0.89, left=0, right=-17)
    >>> a
    0.8899993896484375 [0:-17]
    >>> b = resize(a, 0, -6)
    >>> b
    0.890625 [0:-6]

    >>> c = resize(a, size_res=b)
    >>> c
    0.890625 [0:-6]


    """
    return fix.resize(left_index, right_index, size_res, overflow_style=overflow_style, round_style=round_style)


def left_index(x: Sfix):
    """
    Use this in convertible code

    :return: left bound

    >>> a = Sfix(-0.5, 1, -7)
    >>> left_index(a)
    1

    """
    return x.left


def right_index(x: Sfix):
    """
    Use this in convertible code

    :return: right bound

    >>> a = Sfix(-0.5, 1, -7)
    >>> right_index(a)
    -7
    """
    return x.right


def scalb(x: Sfix, i: int):
    n = 2 ** i
    return Sfix(x.val * n, x.left + i, x.right + i)



"""
0x1 - skip MODEl
0x2 - skip HW_MODEL
0x4 - skip RTL
0x8 - skip GATE

can combine, 0x3 - skip MODEL and HW_MODEL
"""
import os
from contextlib import suppress

SKIP_ABOVE_HW_MODEL = 0x4 + 0x8
SKIP_ABOVE_RTL = 0x8

SKIP_SIMULATIONS_MASK = 0x0

# SKIP_SIMULATIONS_MASK = SKIP_ABOVE_RTL
# SKIP_SIMULATIONS_MASK = SKIP_ABOVE_HW_MODEL

with suppress(KeyError):
    f = int(os.environ['SKIP_SIMULATIONS_MASK'], 16)
    SKIP_SIMULATIONS_MASK = f



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from pip.req import parse_requirements
from setuptools import setup, find_packages

# read the docs dont have 3.6 lol...
# import sys
# if sys.version_info < (3, 6):
#     sys.exit('Sorry, Python < 3.6 is not supported')


with open('README.rst') as readme_file:
    readme = readme_file.read()


# parse_requirements() returns generator of pip.req.InstallRequirement objects
install_reqs = parse_requirements('requirements.txt', session=False)
requirements = [str(ir.req) for ir in install_reqs]

test_requirements = [

]

setup(
    name='pyha',
    version='0.0.1',
    description="Pyha",
    long_description=readme,
    author="Gaspar Karm",
    author_email='gkarm@live.com',
    url='https://github.com/petspats/pyha',

    # package_dir={'':'pyha'},
    packages=find_packages(),
    package_data={'pyha': ['common/hdl/pyha_util.vhd', 'common/hdl/fixed_pkg/*']},
    # py_modules=["pyha"],

    include_package_data=False,
    install_requires=requirements,
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords='pyha',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements,
)



import numpy as np

from pyha.common.hwsim import HW
from pyha.common.sfix import resize, ComplexSfix, Sfix, fixed_truncate


class Conjugate(HW):
    def __init__(self):
        self.outreg = ComplexSfix()

        self._delay = 1

    def main(self, x):
        self.next.outreg.real = x.real
        self.next.outreg.imag = resize(-x.imag, size_res=x.imag)
        return self.outreg

    def model_main(self, x):
        return np.conjugate(x)


class ComplexMultiply(HW):
    """ (x + yj)(u + vj) = (xu - yv) + (xv + yu)j """

    def __init__(self):
        self.real_xu = Sfix()
        self.real_yv = Sfix()

        self.imag_xv = Sfix()
        self.imag_yu = Sfix()

        self.outreg = ComplexSfix()

        self._delay = 2

    def main(self, a, b):
        # assert a.has_same_bounds(b)
        # self.next.real_xu = a.real * b.real
        # self.next.real_yv = a.imag * b.imag
        # self.next.outreg.real = resize(self.real_xu - self.real_yv,
        #                                size_res=a.real, round_style=fixed_truncate)
        #
        # self.next.imag_xv = a.real * b.imag
        # self.next.imag_yu = a.imag * b.real
        # self.next.outreg.imag = resize(self.imag_xv + self.imag_yu,
        #                                size_res=a.imag, round_style=fixed_truncate)

        self.next.real_xu = resize(a.real * b.real, size_res=a.real, round_style=fixed_truncate)
        self.next.real_yv = resize(a.imag * b.imag, size_res=a.real, round_style=fixed_truncate)
        self.next.outreg.real = resize(self.real_xu - self.real_yv,
                                       size_res=a.real, round_style=fixed_truncate)

        self.next.imag_xv = resize(a.real * b.imag, size_res=a.real, round_style=fixed_truncate)
        self.next.imag_yu = resize(a.imag * b.real, size_res=a.real, round_style=fixed_truncate)
        self.next.outreg.imag = resize(self.imag_xv + self.imag_yu,
                                       size_res=a.real, round_style=fixed_truncate)

        return self.outreg

    def model_main(self, a, b):
        return np.array(a) * np.array(b)



from enum import Enum

import numpy as np

from pyha.common.const import Const
from pyha.common.hwsim import HW
from pyha.common.sfix import resize, Sfix, right_index, left_index, fixed_wrap, fixed_truncate, ComplexSfix


# Decent cordic overview
# readable paper -> http://www.andraka.com/files/crdcsrvy.pdf
# vhdl implementation -> https://github.com/Nuand/bladeRF/blob/master/hdl/fpga/ip/nuand/synthesis/cordic.vhd

class CordicMode(Enum):
    VECTORING, ROTATION = range(2)


class Cordic(HW):
    """
    CORDIC algorithm.

    readable paper -> http://www.andraka.com/files/crdcsrvy.pdf

    :param iterations: resource/ precision trade off
    :param mode: vectoring or rotation
    """

    def __init__(self, iterations, mode):
        self.mode = Const(mode)
        self.iterations = iterations

        # + 1 is basically for initial step registers it also helps pipeline code
        self.iterations = iterations + 1
        self.phase_lut = [float(np.arctan(2 ** -i) / np.pi) for i in range(self.iterations)]
        self.phase_lut_fix = Const([Sfix(x, 0, -24) for x in self.phase_lut])

        # todo: this would fail GATE simulation, no errors, quick RTL inspection looks OK
        # self.phase_lut_fix = Const(self.phase_lut)

        # pipeline registers
        self.x = [Sfix()] * self.iterations
        self.y = [Sfix()] * self.iterations
        self.phase = [Sfix()] * self.iterations

        self._delay = self.iterations

    def main(self, x, y, phase):
        """
        Runs one step of pipelined CORDIC
        Returned phase is in 1 to -1 range
        """
        self.initial_step(phase, x, y)

        for i in range(len(self.phase_lut_fix) - 1):
            self.next.x[i + 1], self.next.y[i + 1], self.next.phase[i + 1] = \
                self.pipeline_step(i, self.x[i], self.y[i], self.phase[i], self.phase_lut_fix[i])

        return self.x[-1], self.y[-1], self.phase[-1]

    def initial_step(self, phase, x, y):
        """
        CORDIC works in only 1 quadrant, this performs hacks to make it usable on other qudrants as well.
        """
        self.next.x[0] = x
        self.next.y[0] = y
        self.next.phase[0] = phase
        if self.mode == CordicMode.ROTATION:
            if phase > 0.5:
                # > np.pi/2
                self.next.x[0] = resize(-x, size_res=x)
                self.next.phase[0] = resize(phase - 1.0, size_res=phase)
            elif phase < -0.5:
                # < -np.pi/2
                self.next.x[0] = resize(-x, size_res=x)
                self.next.phase[0] = resize(phase + 1.0, size_res=phase)

        elif self.mode == CordicMode.VECTORING:
            if x < 0.0 and y > 0.0:
                # vector in II quadrant -> initial shift by PI to IV quadrant (mirror)
                self.next.x[0] = resize(-x, size_res=x)
                self.next.y[0] = resize(-y, size_res=y)
                self.next.phase[0] = Sfix(1.0, phase)
            elif x < 0.0 and y < 0.0:
                # vector in III quadrant -> initial shift by -PI to I quadrant (mirror)
                self.next.x[0] = resize(-x, size_res=x)
                self.next.y[0] = resize(-y, size_res=y)
                self.next.phase[0] = Sfix(-1.0, phase)

    def pipeline_step(self, i, x, y, p, adj):
        """
        Core combinatory part, this will be pipeline

        :param i: current pipeline step
        :param adj: phase adjustment
        """
        if self.mode == CordicMode.ROTATION:
            direction = p > 0
        elif self.mode == CordicMode.VECTORING:
            direction = y < 0

        if direction:
            next_x = resize(x - (y >> i), size_res=x, overflow_style=fixed_wrap)
            next_y = resize(y + (x >> i), size_res=y, overflow_style=fixed_wrap)
            next_phase = resize(p - adj, size_res=p, overflow_style=fixed_wrap)
        else:
            next_x = resize(x + (y >> i), size_res=x, overflow_style=fixed_wrap)
            next_y = resize(y - (x >> i), size_res=y, overflow_style=fixed_wrap)
            next_phase = resize(p + adj, size_res=p, overflow_style=fixed_wrap)
        return next_x, next_y, next_phase


    def model_main(self, x, y, phase):
        # this model uses (y * (2 ** -i)) for shift right. meaning it loses no precision for this operation
        # for that reason model and hw_model will not be perfectly matched, can expect 1e-4 to 1e-5 rtol/atol
        # actually not 100% sure that is the reason for strange behaviour
        def cord_model(x, y, phase):
            if self.mode == CordicMode.ROTATION:
                if phase > 0.5:
                    x = -x
                    phase -= 1.0
                elif phase < -0.5:
                    x = -x
                    phase += 1.0
            elif self.mode == CordicMode.VECTORING:
                if x < 0.0 and y > 0.0:
                    # vector in II quadrant -> initial shift by PI to IV quadrant (mirror)
                    x = -x
                    y = -y
                    phase = 1.0
                elif x < 0.0 and y < 0.0:
                    # vector in III quadrant -> initial shift by -PI to I quadrant (mirror)
                    x = -x
                    y = -y
                    phase = -1.0

            for i, adj in enumerate(self.phase_lut):
                if self.mode == CordicMode.ROTATION:
                    sign = 1 if phase > 0 else -1
                elif self.mode == CordicMode.VECTORING:
                    sign = 1 if y < 0 else -1
                x, y, phase = x - sign * (y * (2 ** -i)), y + sign * (x * (2 ** -i)), phase - sign * adj
            return x, y, phase

        return [cord_model(xx, yy, pp) for xx, yy, pp in zip(x, y, phase)]


class NCO(HW):
    """
    Signal generator. Input phase must be normalized to -1 to 1 range!

    :param cordic_iterations: performace/resource usage trade off
    """
    def __init__(self, cordic_iterations=16):

        self.cordic = Cordic(cordic_iterations, CordicMode.ROTATION)
        self.phase_acc = Sfix()
        self._delay = self.cordic.iterations + 1

    def main(self, phase_inc):
        """
        18 bits precision.

        :param phase_inc: amount of rotation applied for next clock cycle, must be normalized to -1 to 1.
        :return: baseband signal
        :rtype: ComplexSfix
        """
        self.next.phase_acc = resize(self.phase_acc + phase_inc, size_res=phase_inc, overflow_style=fixed_wrap,
                                     round_style=fixed_truncate)

        start_x = Sfix(1.0 / 1.646760, 1, -17)  # gets rid of cordic gain, could add amplitude modulation here
        start_y = Sfix(0.0, 1, -17)  # 1 bit for gains, remove later

        x, y, phase = self.cordic.main(start_x, start_y, self.phase_acc)
        xr = resize(x, 0, -17)
        yr = resize(y, 0, -17)
        retc = ComplexSfix(xr, yr)
        return retc

    def model_main(self, phase_list):
        p = np.cumsum(phase_list * np.pi)
        return np.exp(p * 1j)


class ToPolar(HW):
    """
    Converts IQ to polar.
    """

    def __init__(self):
        self.core = Cordic(13, CordicMode.VECTORING)
        self.out_abs = Sfix()
        self.out_angle = Sfix()

        self._delay = self.core.iterations + 1

    def main(self, c):
        """
        :param c: baseband in, internal sizes are derived from this.
        :type c: ComplexSfix
        :return: abs (gain corrected) angle (in 1 to -1 range)
        """
        phase = Sfix(0.0, 0, -24)

        # give 1 extra bit, as there is stuff like CORDIC gain.. in some cases 2 bits may be needed!
        # there will be CORDIC gain + abs value held by x can be > 1
        # remove 1 bit from fractional part, to keep 18 bit numbers
        x = resize(c.real, left_index(c.real) + 1, right_index(c.real) + 1, round_style=fixed_truncate)
        y = resize(c.imag, left_index(c.imag) + 1, right_index(c.imag) + 1, round_style=fixed_truncate)

        abs, _, angle = self.core.main(x, y, phase)

        # get rid of CORDIC gain and extra bits
        self.next.out_abs = resize(abs * (1.0 / 1.646760), c.imag, round_style=fixed_truncate)
        self.next.out_angle = resize(angle, c.imag, round_style=fixed_truncate)
        return self.out_abs, self.out_angle

    def model_main(self, cin):
        # note that angle in -1..1 range
        rabs = [np.abs(x) for x in cin]
        angle = [np.angle(x) / np.pi for x in cin]
        return rabs, angle


class Angle(HW):
    """
    Wrapper around ToPolar. Abs output will be optimized away.
    """
    def __init__(self):
        self.core = ToPolar()
        self._delay = self.core._delay

    def main(self, c):
        _, angle = self.core.main(c)
        return angle

    def model_main(self, cin):
        # note that angle in -1..1 range
        return [np.angle(x) / np.pi for x in cin]


class Abs(HW):
    """
    Wrapper around ToPolar. Angle output will be optimized away.
    """
    def __init__(self):
        self.core = ToPolar()
        self._delay = self.core._delay

    def main(self, c):
        abs, _ = self.core.main(c)
        return abs

    def model_main(self, cin):
        # note that angle in -1..1 range
        return [np.abs(x) for x in cin]
